from __future__ import division  # confidence high
from __future__ import with_statement

import itertools
import warnings

import numpy as np

import pyfits

from pyfits.card import _pad
from pyfits.util import encode_ascii
from pyfits.tests import PyfitsTestCase
from pyfits.tests.util import catch_warnings, ignore_warnings, CaptureStdio

from nose.tools import assert_equal, assert_false, assert_raises, assert_true


class TestOldApiHeaderFunctions(PyfitsTestCase):
    """
    Tests that specifically use attributes and methods from the old
    Header/CardList API from PyFITS 3.0 and prior.

    This tests backward compatibility support for those interfaces.
    """

    def test_ascardimage_verifies_the_comment_string_to_be_ascii_text(self):
        # the ascardimage() verifies the comment string to be ASCII text
        c = pyfits.Card.fromstring('abc     = +  2.1   e + 12 / abcde\0')
        assert_raises(Exception, c.ascardimage)

    def test_rename_key(self):
        """Test backwards compatibility support for Header.rename_key()"""
        header = pyfits.Header([('A', 'B', 'C'), ('D', 'E', 'F')])
        header.rename_key('A', 'B')
        assert_true('A' not in header)
        assert_true('B' in header)
        assert_equal(header[0], 'B')
        assert_equal(header['B'], 'B')
        assert_equal(header.comments['B'], 'C')

    def test_add_commentary(self):
        header = pyfits.Header([('A', 'B', 'C'), ('HISTORY', 1),
                                ('HISTORY', 2), ('HISTORY', 3), ('', '', ''),
                                ('', '', '')])
        header.add_history(4)
        # One of the blanks should get used, so the length shouldn't change
        assert_equal(len(header), 6)
        assert_equal(header.cards[4].value, 4)
        assert_equal(header['HISTORY'], [1, 2, 3, 4])

        header.add_history(0, after='A')
        assert_equal(len(header), 6)
        assert_equal(header.cards[1].value, 0)
        assert_equal(header['HISTORY'], [0, 1, 2, 3, 4])

        header = pyfits.Header([('A', 'B', 'C'), ('', 1), ('', 2), ('', 3),
                                ('', '', ''), ('', '', '')])
        header.add_blank(4)
        # This time a new blank should be added, and the existing blanks don't
        # get used... (though this is really kinda sketchy--there's a
        # distinction between truly blank cards, and cards with blank keywords
        # that isn't currently made int he code)
        assert_equal(len(header), 7)
        assert_equal(header.cards[6].value, 4)
        assert_equal(header[''], [1, 2, 3, '', '', 4])

        header.add_blank(0, after='A')
        assert_equal(len(header), 8)
        assert_equal(header.cards[1].value, 0)
        assert_equal(header[''], [0, 1, 2, 3, '', '', 4])

    def test_has_key(self):
        header = pyfits.Header([('A', 'B', 'C'), ('D', 'E', 'F')])
        assert_true(header.has_key('A'))
        assert_true(header.has_key('D'))
        assert_false(header.has_key('C'))

    def test_totxtfile(self):
        hdul = pyfits.open(self.data('test0.fits'))
        hdul[0].header.toTxtFile(self.temp('header.txt'))
        hdu = pyfits.ImageHDU()
        hdu.header.update('MYKEY', 'FOO', 'BAR')
        hdu.header.fromTxtFile(self.temp('header.txt'), replace=True)
        assert_equal(len(hdul[0].header.ascard), len(hdu.header.ascard))
        assert_false(hdu.header.has_key('MYKEY'))
        assert_false(hdu.header.has_key('EXTENSION'))
        assert_true(hdu.header.has_key('SIMPLE'))

        # Write the hdu out and read it back in again--it should be recognized
        # as a PrimaryHDU
        hdu.writeto(self.temp('test.fits'), output_verify='ignore')
        assert_true(isinstance(pyfits.open(self.temp('test.fits'))[0],
                               pyfits.PrimaryHDU))

        hdu = pyfits.ImageHDU()
        hdu.header.update('MYKEY', 'FOO', 'BAR')
        hdu.header.fromTxtFile(self.temp('header.txt'))
        # hdu.header should have MYKEY keyword, and also adds PCOUNT and
        # GCOUNT, giving it 3 more keywords in total than the original
        assert_equal(len(hdul[0].header.ascard), len(hdu.header.ascard) - 3)
        assert_true(hdu.header.has_key('MYKEY'))
        assert_false(hdu.header.has_key('EXTENSION'))
        assert_true(hdu.header.has_key('SIMPLE'))

        with ignore_warnings():
            hdu.writeto(self.temp('test.fits'), output_verify='ignore',
                        clobber=True)
        hdul2 = pyfits.open(self.temp('test.fits'))
        assert_true(len(hdul2), 2)
        assert_true(hdul2[1].header.has_key('MYKEY'))

    def test_update_comment(self):
        hdul = pyfits.open(self.data('arange.fits'))
        hdul[0].header.update('FOO', 'BAR', 'BAZ')
        assert_equal(hdul[0].header['FOO'], 'BAR')
        assert_equal(hdul[0].header.ascard['FOO'].comment, 'BAZ')

        hdul.writeto(self.temp('test.fits'))

        hdul = pyfits.open(self.temp('test.fits'), mode='update')
        hdul[0].header.ascard['FOO'].comment = 'QUX'
        hdul.close()

        hdul = pyfits.open(self.temp('test.fits'))
        assert_equal(hdul[0].header.ascard['FOO'].comment, 'QUX')

    def test_long_commentary_card(self):
        # Another version of this test using new API methods is found in
        # TestHeaderFunctions
        header = pyfits.Header()
        header.update('FOO', 'BAR')
        header.update('BAZ', 'QUX')
        longval = 'ABC' * 30
        header.add_history(longval)
        header.update('FRED', 'BARNEY')
        header.add_history(longval)

        assert_equal(len(header.ascard), 7)
        assert_equal(header.ascard[2].key, 'FRED')
        assert_equal(str(header.cards[3]), 'HISTORY ' + longval[:72])
        assert_equal(str(header.cards[4]).rstrip(), 'HISTORY ' + longval[72:])

        header.add_history(longval, after='FOO')
        assert_equal(len(header.ascard), 9)
        assert_equal(str(header.cards[1]), 'HISTORY ' + longval[:72])
        assert_equal(str(header.cards[2]).rstrip(), 'HISTORY ' + longval[72:])

    def test_wildcard_slice(self):
        """Test selecting a subsection of a header via wildcard matching."""

        header = pyfits.Header()
        header.update('ABC', 0)
        header.update('DEF', 1)
        header.update('ABD', 2)
        cards = header.ascard['AB*']
        assert_equal(len(cards), 2)
        assert_equal(cards[0].value, 0)
        assert_equal(cards[1].value, 2)

    def test_assign_boolean(self):
        """
        Regression test for #123. Tests assigning Python and Numpy boolean
        values to keyword values.
        """

        fooimg = _pad('FOO     =                    T')
        barimg = _pad('BAR     =                    F')
        h = pyfits.Header()
        h.update('FOO', True)
        h.update('BAR', False)
        assert_equal(h['FOO'], True)
        assert_equal(h['BAR'], False)
        assert_equal(h.ascard['FOO'].cardimage, fooimg)
        assert_equal(h.ascard['BAR'].cardimage, barimg)

        h = pyfits.Header()
        h.update('FOO', np.bool_(True))
        h.update('BAR', np.bool_(False))
        assert_equal(h['FOO'], True)
        assert_equal(h['BAR'], False)
        assert_equal(h.ascard['FOO'].cardimage, fooimg)
        assert_equal(h.ascard['BAR'].cardimage, barimg)

        h = pyfits.Header()
        h.ascard.append(pyfits.Card.fromstring(fooimg))
        h.ascard.append(pyfits.Card.fromstring(barimg))
        assert_equal(h['FOO'], True)
        assert_equal(h['BAR'], False)
        assert_equal(h.ascard['FOO'].cardimage, fooimg)
        assert_equal(h.ascard['BAR'].cardimage, barimg)

    def test_cardlist_list_methods(self):
        """Regression test for #190."""

        header = pyfits.Header()
        header.update('A', 'B', 'C')
        header.update('D', 'E', 'F')
        # The old header.update method won't let you append a duplicate keyword
        header.append(('D', 'G', 'H'))

        assert_equal(header.ascardlist().index(header.cards['A']), 0)
        assert_equal(header.ascardlist().index(header.cards['D']), 1)
        assert_equal(header.ascardlist().index(header.cards[('D', 1)]), 2)

        # Since the original CardList class really only works on card objects
        # the count method is mostly useless since cards didn't used to compare
        # equal sensibly
        assert_equal(header.ascardlist().count(header.cards['A']), 1)
        assert_equal(header.ascardlist().count(header.cards['D']), 1)
        assert_equal(header.ascardlist().count(header.cards[('D', 1)]), 1)
        assert_equal(header.ascardlist().count(pyfits.Card('A', 'B', 'C')), 0)


class TestHeaderFunctions(PyfitsTestCase):
    """Test PyFITS Header and Card objects."""

    def test_card_constructor_default_args(self):
        """Test Card constructor with default argument values."""

        c = pyfits.Card()
        assert_equal('', c.key)

    def test_string_value_card(self):
        """Test Card constructor with string value"""

        c = pyfits.Card('abc', '<8 ch')
        assert_equal(str(c),
                     "ABC     = '<8 ch   '                                                            ")
        c = pyfits.Card('nullstr', '')
        assert_equal(str(c),
                     "NULLSTR = ''                                                                    ")

    def test_boolean_value_card(self):
        """Test Card constructor with boolean value"""

        c = pyfits.Card("abc", True)
        assert_equal(str(c),
                     "ABC     =                    T                                                  ")

        c = pyfits.Card.fromstring('abc     = F')
        assert_equal(c.value, False)

    def test_long_integer_value_card(self):
        """Test Card constructor with long integer value"""

        c = pyfits.Card('long_int', -467374636747637647347374734737437)
        assert_equal(str(c),
                     "LONG_INT= -467374636747637647347374734737437                                    ")

    def test_floating_point_value_card(self):
        """Test Card constructor with floating point value"""

        c = pyfits.Card('floatnum', -467374636747637647347374734737437.)

        if (str(c) != "FLOATNUM= -4.6737463674763E+32                                                  " and
            str(c) != "FLOATNUM= -4.6737463674763E+032                                                 "):
            assert_equal(str(c),
                         "FLOATNUM= -4.6737463674763E+32                                                  ")

    def test_complex_value_card(self):
        """Test Card constructor with complex value"""

        c = pyfits.Card('abc',
                        1.2345377437887837487e88+6324767364763746367e-33j)

        if (str(c) != "ABC     = (1.23453774378878E+88, 6.32476736476374E-15)                          " and
            str(c) != "ABC     = (1.2345377437887E+088, 6.3247673647637E-015)                          "):
            assert_equal(str(c),
                         "ABC     = (1.23453774378878E+88, 6.32476736476374E-15)                          ")

    def test_card_image_constructed_too_long(self):
        """Test that over-long cards truncate the comment"""

        # card image constructed from key/value/comment is too long
        # (non-string value)
        with ignore_warnings():
            c = pyfits.Card('abc', 9, 'abcde'*20)
            assert_equal(str(c),
                         "ABC     =                    9 / abcdeabcdeabcdeabcdeabcdeabcdeabcdeabcdeabcdeab")
            c = pyfits.Card('abc', 'a'*68, 'abcdefg')
            assert_equal(str(c),
                         "ABC     = 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'")

    def test_constructor_filter_illegal_data_structures(self):
        """Test that Card constructor raises exceptions on bad arguments"""

        assert_raises(ValueError, pyfits.Card, ('abc',), {'value': (2, 3)})
        assert_raises(ValueError, pyfits.Card, 'key', [], 'comment')

    def test_keyword_too_long(self):
        """Test that long Card keywords are allowed, but with a warning"""

        with catch_warnings():
            warnings.simplefilter('error')
            assert_raises(UserWarning, pyfits.Card, 'abcdefghi', 'long')

    def test_illegal_characters_in_key(self):
        """
        Test that Card constructor allows illegal characters in the keyword,
        but creates a HIERARCH card.
        """

        # This test used to check that a ValueError was raised, because a
        # keyword like 'abc+' was simply not allowed.  Now it should create a
        # HIERARCH card.

        with catch_warnings(record=True) as w:
            c = pyfits.Card('abc+', 9)
            assert_equal(len(w), 1)
            assert_equal(c.image, _pad('HIERARCH abc+ =                    9'))

    def test_commentary_cards(self):
        # commentary cards
        c = pyfits.Card("history",
                        "A commentary card's value has no quotes around it.")
        assert_equal(str(c),
                     "HISTORY A commentary card's value has no quotes around it.                      ")
        c = pyfits.Card("comment",
                        "A commentary card has no comment.", "comment")
        assert_equal(str(c),
                     "COMMENT A commentary card has no comment.                                       ")

    def test_commentary_card_created_by_fromstring(self):
        # commentary card created by fromstring()
        c = pyfits.Card.fromstring("COMMENT card has no comments. / text after slash is still part of the value.")
        assert_equal(c.value,
                     'card has no comments. / text after slash is still part of the value.')
        assert_equal(c.comment, '')

    def test_commentary_card_will_not_parse_numerical_value(self):
        # commentary card will not parse the numerical value
        c = pyfits.Card.fromstring("history  (1, 2)")
        assert_equal(str(c),
                     "HISTORY  (1, 2)                                                                 ")

    def test_equal_sign_after_column8(self):
        # equal sign after column 8 of a commentary card will be part ofthe
        # string value
        c = pyfits.Card.fromstring("history =   (1, 2)")
        assert_equal(str(c),
                     "HISTORY =   (1, 2)                                                              ")

    def test_blank_keyword(self):
        c = pyfits.Card('', '       / EXPOSURE INFORMATION')
        assert_equal(str(c),
                     '               / EXPOSURE INFORMATION                                           ')
        c = pyfits.Card.fromstring(str(c))
        assert_equal(c.keyword, '')
        assert_equal(c.value, '       / EXPOSURE INFORMATION')


    def test_specify_undefined_value(self):
        # this is how to specify an undefined value
        c = pyfits.Card("undef", pyfits.card.UNDEFINED)
        assert_equal(str(c),
                     "UNDEF   =                                                                       ")

    def test_complex_number_using_string_input(self):
        # complex number using string input
        c = pyfits.Card.fromstring('abc     = (8, 9)')
        assert_equal(str(c),
                     "ABC     =               (8, 9)                                                  ")

    def test_fixable_non_standard_fits_card(self):
        # fixable non-standard FITS card will keep the original format
        c = pyfits.Card.fromstring('abc     = +  2.1   e + 12')
        assert_equal(c.value, 2100000000000.0)
        with CaptureStdio():
            assert_equal(str(c),
                         "ABC     =             +2.1E+12                                                  ")

    def test_fixable_non_fsc(self):
        # fixable non-FSC: if the card is not parsable, it's value will be
        # assumed
        # to be a string and everything after the first slash will be comment
        c = pyfits.Card.fromstring("no_quote=  this card's value has no quotes / let's also try the comment")
        with CaptureStdio():
            assert_equal(str(c),
                         "NO_QUOTE= 'this card''s value has no quotes' / let's also try the comment       ")

    def test_undefined_value_using_string_input(self):
        # undefined value using string input
        c = pyfits.Card.fromstring('abc     =    ')
        assert_equal(str(c),
                     "ABC     =                                                                       ")

    def test_mislocated_equal_sign(self):
        # test mislocated "=" sign
        c = pyfits.Card.fromstring('xyz= 100')
        assert_equal(c.keyword, 'XYZ')
        assert_equal(c.value, 100)
        with CaptureStdio():
            assert_equal(str(c),
                         "XYZ     =                  100                                                  ")

    def test_equal_only_up_to_column_10(self):
        # the test of "=" location is only up to column 10

        # This test used to check if PyFITS rewrote this card to a new format,
        # something like "HISTO   = '=   (1, 2)".  But since ticket #109 if the
        # format is completely wrong we don't make any assumptions and the card
        # should be left alone
        with CaptureStdio():
            c = pyfits.Card.fromstring("histo       =   (1, 2)")
            assert_equal(str(c), _pad("histo       =   (1, 2)"))

            # Likewise this card should just be left in its original form and
            # we shouldn't guess how to parse it or rewrite it.
            c = pyfits.Card.fromstring("   history          (1, 2)")
            assert_equal(str(c), _pad("   history          (1, 2)"))

    def test_verify_invalid_equal_sign(self):
        # verification
        c = pyfits.Card.fromstring('abc= a6')
        with catch_warnings(record=True) as w:
            with CaptureStdio():
                c.verify()
            err_text1 = ("Card 'ABC' is not FITS standard (equal sign not at "
                         "column 8)")
            err_text2 = ("Card 'ABC' is not FITS standard (invalid value "
                         "string: a6")
            assert_equal(len(w), 4)
            assert_true(err_text1 in str(w[1].message))
            assert_true(err_text2 in str(w[2].message))

    def test_fix_invalid_equal_sign(self):
        c = pyfits.Card.fromstring('abc= a6')
        with catch_warnings(record=True) as w:
            with CaptureStdio():
                c.verify('fix')
            fix_text = "Fixed 'ABC' card to meet the FITS standard."
            assert_equal(len(w), 4)
            assert_true(fix_text in str(w[1].message))
        assert_equal(str(c),
                     "ABC     = 'a6      '                                                            ")

    def test_long_string_value(self):
        # test long string value
        c = pyfits.Card('abc', 'long string value '*10, 'long comment '*10)
        assert_equal(str(c),
            "ABC     = 'long string value long string value long string value long string &' "
            "CONTINUE  'value long string value long string value long string value long &'  "
            "CONTINUE  'string value long string value long string value &'                  "
            "CONTINUE  '&' / long comment long comment long comment long comment long        "
            "CONTINUE  '&' / comment long comment long comment long comment long comment     "
            "CONTINUE  '&' / long comment                                                    ")

    def test_long_unicode_string(self):
        """Regression test for
        https://github.com/spacetelescope/PyFITS/issues/1

        So long as a unicode string can be converted to ASCII it should have no
        different behavior in this regard from a byte string.
        """

        h1 = pyfits.Header()
        h1['TEST'] = 'abcdefg' * 30

        h2 = pyfits.Header()
        with catch_warnings(record=True) as w:
            h2['TEST'] = u'abcdefg' * 30
            assert_equal(len(w), 0)

        assert_equal(str(h1), str(h2))

    def test_long_string_repr(self):
        """Regression test for #193

        Ensure that the __repr__() for cards represented with CONTINUE cards is
        split across multiple lines (broken at each *physical* card).
        """

        header = pyfits.Header()
        header['TEST1'] = ('Regular value', 'Regular comment')
        header['TEST2'] = ('long string value ' * 10, 'long comment ' * 10)
        header['TEST3'] = ('Regular value', 'Regular comment')

        assert_equal(repr(header).splitlines(),
            [str(pyfits.Card('TEST1', 'Regular value', 'Regular comment')),
             "TEST2   = 'long string value long string value long string value long string &' ",
             "CONTINUE  'value long string value long string value long string value long &'  ",
             "CONTINUE  'string value long string value long string value &'                  ",
             "CONTINUE  '&' / long comment long comment long comment long comment long        ",
             "CONTINUE  '&' / comment long comment long comment long comment long comment     ",
             "CONTINUE  '&' / long comment                                                    ",
             str(pyfits.Card('TEST3', 'Regular value', 'Regular comment'))])

    def test_blank_keyword_long_value(self):
        """Regression test for #194

        Test that a blank keyword ('') can be assigned a too-long value that is
        continued across multiple cards with blank keywords, just like COMMENT
        and HISTORY cards.
        """

        value = 'long string value ' * 10
        header = pyfits.Header()
        header[''] = value

        assert_equal(len(header), 3)
        assert_equal(' '.join(header['']), value.rstrip())

        # Ensure that this works like other commentary keywords
        header['COMMENT'] = value
        header['HISTORY'] = value
        assert_equal(header['COMMENT'], header['HISTORY'])
        assert_equal(header['COMMENT'], header[''])

    def test_long_string_from_file(self):
        c = pyfits.Card('abc', 'long string value '*10, 'long comment '*10)
        hdu = pyfits.PrimaryHDU()
        hdu.header.append(c)
        hdu.writeto(self.temp('test_new.fits'))

        hdul = pyfits.open(self.temp('test_new.fits'))
        c = hdul[0].header.cards['abc']
        hdul.close()
        assert_equal(str(c),
            "ABC     = 'long string value long string value long string value long string &' "
            "CONTINUE  'value long string value long string value long string value long &'  "
            "CONTINUE  'string value long string value long string value &'                  "
            "CONTINUE  '&' / long comment long comment long comment long comment long        "
            "CONTINUE  '&' / comment long comment long comment long comment long comment     "
            "CONTINUE  '&' / long comment                                                    ")


    def test_word_in_long_string_too_long(self):
        # if a word in a long string is too long, it will be cut in the middle
        c = pyfits.Card('abc', 'longstringvalue'*10, 'longcomment'*10)
        assert_equal(str(c),
            "ABC     = 'longstringvaluelongstringvaluelongstringvaluelongstringvaluelongstr&'"
            "CONTINUE  'ingvaluelongstringvaluelongstringvaluelongstringvaluelongstringvalu&'"
            "CONTINUE  'elongstringvalue&'                                                   "
            "CONTINUE  '&' / longcommentlongcommentlongcommentlongcommentlongcommentlongcomme"
            "CONTINUE  '&' / ntlongcommentlongcommentlongcommentlongcomment                  ")

    def test_long_string_value_via_fromstring(self):
        # long string value via fromstring() method
        c = pyfits.Card.fromstring(
            _pad("abc     = 'longstring''s testing  &  ' / comments in line 1") +
            _pad("continue  'continue with long string but without the ampersand at the end' /") +
            _pad("continue  'continue must have string value (with quotes)' / comments with ''. "))
        assert_equal(str(c),
            "ABC     = 'longstring''s testing  continue with long string but without the &'  "
            "CONTINUE  'ampersand at the endcontinue must have string value (with quotes)&'  "
            "CONTINUE  '&' / comments in line 1 comments with ''.                            ")

    def test_continue_card_with_equals_in_value(self):
        """
        Regression test for #117.
        """

        c = pyfits.Card.fromstring(
            _pad("EXPR    = '/grp/hst/cdbs//grid/pickles/dat_uvk/pickles_uk_10.fits * &'") +
            _pad("CONTINUE  '5.87359e-12 * MWAvg(Av=0.12)&'") +
            _pad("CONTINUE  '&' / pysyn expression"))

        assert_equal(c.keyword, 'EXPR')
        assert_equal(c.value,
                     '/grp/hst/cdbs//grid/pickles/dat_uvk/pickles_uk_10.fits '
                     '* 5.87359e-12 * MWAvg(Av=0.12)')
        assert_equal(c.comment, 'pysyn expression')

    def test_hierarch_card_creation(self):
        # Test automatic upgrade to hierarch card
        with catch_warnings(record=True) as w:
            c = pyfits.Card('ESO INS SLIT2 Y1FRML',
                            'ENC=OFFSET+RESOL*acos((WID-(MAX+MIN))/(MAX-MIN)')
            assert_equal(len(w), 1)
            assert_true('HIERARCH card will be created' in str(w[0].message))
            assert_equal(str(c),
                         "HIERARCH ESO INS SLIT2 Y1FRML= "
                         "'ENC=OFFSET+RESOL*acos((WID-(MAX+MIN))/(MAX-MIN)'")

        # Test manual creation of hierarch card
        c = pyfits.Card('hierarch abcdefghi', 10)
        assert_equal(str(c),
            "HIERARCH abcdefghi = "
            "10                                                         ")
        c = pyfits.Card('HIERARCH ESO INS SLIT2 Y1FRML',
                        'ENC=OFFSET+RESOL*acos((WID-(MAX+MIN))/(MAX-MIN)')
        assert_equal(str(c),
                     "HIERARCH ESO INS SLIT2 Y1FRML= "
                     "'ENC=OFFSET+RESOL*acos((WID-(MAX+MIN))/(MAX-MIN)'")

    def test_hierarch_with_abbrev_value_indicator(self):
        """Regression test for
        https://github.com/spacetelescope/PyFITS/issues/5
        """

        c = pyfits.Card.fromstring("HIERARCH key.META_4='calFileVersion'")
        assert_equal(c.keyword, 'key.META_4')
        assert_equal(c.value, 'calFileVersion')
        assert_equal(c.comment, '')

    def test_hierarch_keyword_whitespace(self):
        """
        Regression test for
        https://github.com/spacetelescope/PyFITS/issues/6

        Make sure any leading or trailing whitespace around HIERARCH
        keywords is stripped from the actual keyword value.
        """

        c = pyfits.Card.fromstring(
                "HIERARCH  key.META_4    = 'calFileVersion'")
        assert_equal(c.keyword, 'key.META_4')
        assert_equal(c.value, 'calFileVersion')
        assert_equal(c.comment, '')

        # Test also with creation via the Card constructor
        c = pyfits.Card('HIERARCH  key.META_4', 'calFileVersion')
        assert_equal(c.keyword, 'key.META_4')
        assert_equal(c.value, 'calFileVersion')
        assert_equal(c.comment, '')

    def test_verify_mixed_case_hierarch(self):
        """Regression test for
        https://github.com/spacetelescope/PyFITS/issues/7

        Assures that HIERARCH keywords with lower-case characters and other
        normally invalid keyword characters are not considered invalid.
        """

        c = pyfits.Card('HIERARCH WeirdCard.~!@#_^$%&', 'The value',
                        'a comment')
        # This should not raise any exceptions
        c.verify('exception')
        assert_equal(c.keyword, 'WeirdCard.~!@#_^$%&')
        assert_equal(c.value, 'The value')
        assert_equal(c.comment, 'a comment')

        # Test also the specific case from the original bug report
        header = pyfits.Header([
            ('simple', True),
            ('BITPIX', 8),
            ('NAXIS', 0),
            ('EXTEND', True, 'May contain datasets'),
            ('HIERARCH key.META_0', 'detRow')
        ])
        hdu = pyfits.PrimaryHDU(header=header)
        hdu.writeto(self.temp('test.fits'))
        with pyfits.open(self.temp('test.fits')) as hdul:
            header2 = hdul[0].header
            assert_equal(str(header.cards[header.index('key.META_0')]),
                         str(header2.cards[header2.index('key.META_0')]))

    def test_missing_keyword(self):
        """Test that accessing a non-existent keyword raises a KeyError."""

        header = pyfits.Header()
        assert_raises(KeyError, lambda k: header[k], 'NAXIS')
        # Test the exception message
        try:
            header['NAXIS']
        except KeyError, e:
            assert_equal(e.args[0], "Keyword 'NAXIS' not found.")

    def test_hierarch_card_lookup(self):
        header = pyfits.Header()
        header['hierarch abcdefghi'] = 10
        assert_true('abcdefghi' in header)
        assert_equal(header['abcdefghi'], 10)
        # This used to be assert_false, but per ticket #155 hierarch keywords
        # should be treated case-insensitively when performing lookups
        assert_true('ABCDEFGHI' in header)

    def test_hierarch_create_and_update(self):
        """
        Regression test for #158.  Tests several additional use cases for
        working with HIERARCH cards.
        """

        msg = 'a HIERARCH card will be created'

        header = pyfits.Header()
        with catch_warnings(record=True) as w:
            header.update('HIERARCH BLAH BLAH', 'TESTA')
            assert_equal(len(w), 0)
            assert_true('BLAH BLAH' in header)
            assert_equal(header['BLAH BLAH'], 'TESTA')

            header.update('HIERARCH BLAH BLAH', 'TESTB')
            assert_equal(len(w), 0)
            assert_true(header['BLAH BLAH'], 'TESTB')

            # Update without explicitly stating 'HIERARCH':
            header.update('BLAH BLAH', 'TESTC')
            assert_equal(len(w), 0)
            assert_equal(len(header), 1)
            assert_true(header['BLAH BLAH'], 'TESTC')

            # Test case-insensitivity
            header.update('HIERARCH blah blah', 'TESTD')
            assert_equal(len(w), 0)
            assert_equal(len(header), 1)
            assert_true(header['blah blah'], 'TESTD')

            header.update('blah blah', 'TESTE')
            assert_equal(len(w), 0)
            assert_equal(len(header), 1)
            assert_true(header['blah blah'], 'TESTE')

            # Create a HIERARCH card > 8 characters without explicitly stating
            # 'HIERARCH'
            header.update('BLAH BLAH BLAH', 'TESTA')
            assert_equal(len(w), 1)
            assert_true(msg in str(w[0].message))

            header.update('HIERARCH BLAH BLAH BLAH', 'TESTB')
            assert_equal(len(w), 1)
            assert_true(header['BLAH BLAH BLAH'], 'TESTB')

            # Update without explicitly stating 'HIERARCH':
            header.update('BLAH BLAH BLAH', 'TESTC')
            assert_equal(len(w), 1)
            assert_true(header['BLAH BLAH BLAH'], 'TESTC')

            # Test case-insensitivity
            header.update('HIERARCH blah blah blah', 'TESTD')
            assert_equal(len(w), 1)
            assert_true(header['blah blah blah'], 'TESTD')

            header.update('blah blah blah', 'TESTE')
            assert_equal(len(w), 1)
            assert_true(header['blah blah blah'], 'TESTE')

    def test_short_hierarch_create_and_update(self):
        """
        Regression test for #158.  Tests several additional use cases for
        working with HIERARCH cards, specifically where the keyword is fewer
        than 8 characters, but contains invalid characters such that it can
        only be created as a HIERARCH card.
        """

        msg = 'a HIERARCH card will be created'

        header = pyfits.Header()
        with catch_warnings(record=True) as w:
            header.update('HIERARCH BLA BLA', 'TESTA')
            assert_equal(len(w), 0)
            assert_true('BLA BLA' in header)
            assert_equal(header['BLA BLA'], 'TESTA')

            header.update('HIERARCH BLA BLA', 'TESTB')
            assert_equal(len(w), 0)
            assert_true(header['BLA BLA'], 'TESTB')

            # Update without explicitly stating 'HIERARCH':
            header.update('BLA BLA', 'TESTC')
            assert_equal(len(w), 0)
            assert_true(header['BLA BLA'], 'TESTC')

            # Test case-insensitivity
            header.update('HIERARCH bla bla', 'TESTD')
            assert_equal(len(w), 0)
            assert_equal(len(header), 1)
            assert_true(header['bla bla'], 'TESTD')

            header.update('bla bla', 'TESTE')
            assert_equal(len(w), 0)
            assert_equal(len(header), 1)
            assert_true(header['bla bla'], 'TESTE')

        header = pyfits.Header()
        with catch_warnings(record=True) as w:
            # Create a HIERARCH card containing invalid characters without
            # explicitly stating 'HIERARCH'
            header.update('BLA BLA', 'TESTA')
            assert_equal(len(w), 1)
            assert_true(msg in str(w[0].message))

            header.update('HIERARCH BLA BLA', 'TESTB')
            assert_equal(len(w), 1)
            assert_true(header['BLA BLA'], 'TESTB')

            # Update without explicitly stating 'HIERARCH':
            header.update('BLA BLA', 'TESTC')
            assert_equal(len(w), 1)
            assert_true(header['BLA BLA'], 'TESTC')

            # Test case-insensitivity
            header.update('HIERARCH bla bla', 'TESTD')
            assert_equal(len(w), 1)
            assert_equal(len(header), 1)
            assert_true(header['bla bla'], 'TESTD')

            header.update('bla bla', 'TESTE')
            assert_equal(len(w), 1)
            assert_equal(len(header), 1)
            assert_true(header['bla bla'], 'TESTE')

    def test_header_setitem_invalid(self):
        header = pyfits.Header()
        def test():
            header['FOO'] = ('bar', 'baz', 'qux')
        assert_raises(ValueError, test)

    def test_header_setitem_1tuple(self):
        header = pyfits.Header()
        header['FOO'] = ('BAR',)
        assert_equal(header['FOO'], 'BAR')
        assert_equal(header[0], 'BAR')
        assert_equal(header.comments[0], '')
        assert_equal(header.comments['FOO'], '')

    def test_header_setitem_2tuple(self):
        header = pyfits.Header()
        header['FOO'] = ('BAR', 'BAZ')
        assert_equal(header['FOO'], 'BAR')
        assert_equal(header[0], 'BAR')
        assert_equal(header.comments[0], 'BAZ')
        assert_equal(header.comments['FOO'], 'BAZ')

    def test_header_set_value_to_none(self):
        """
        Setting the value of a card to None should simply give that card a
        blank value.
        """

        header = pyfits.Header()
        header['FOO'] = 'BAR'
        assert_equal(header['FOO'], 'BAR')
        header['FOO'] = None
        assert_equal(header['FOO'], '')

    def test_set_comment_only(self):
        header = pyfits.Header([('A', 'B', 'C')])
        header.set('A', comment='D')
        assert_equal(header['A'], 'B')
        assert_equal(header.comments['A'], 'D')

    def test_header_iter(self):
        header = pyfits.Header([('A', 'B'), ('C', 'D')])
        assert_equal(list(header), ['A', 'C'])

    def test_header_slice(self):
        header = pyfits.Header([('A', 'B'), ('C', 'D'), ('E', 'F')])
        newheader = header[1:]
        assert_equal(len(newheader), 2)
        assert_true('A' not in newheader)
        assert_true('C' in newheader)
        assert_true('E' in newheader)

        newheader = header[::-1]
        assert_equal(len(newheader), 3)
        assert_equal(newheader[0], 'F')
        assert_equal(newheader[1], 'D')
        assert_equal(newheader[2], 'B')

        newheader = header[::2]
        assert_equal(len(newheader), 2)
        assert_true('A' in newheader)
        assert_true('C' not in newheader)
        assert_true('E' in newheader)

    def test_header_slice_assignment(self):
        """
        Assigning to a slice should just assign new values to the cards
        included in the slice.
        """

        header = pyfits.Header([('A', 'B'), ('C', 'D'), ('E', 'F')])

        # Test assigning slice to the same value; this works similarly to numpy
        # arrays
        header[1:] = 1
        assert_equal(header[1], 1)
        assert_equal(header[2], 1)

        # Though strings are iterable they should be treated as a scalar value
        header[1:] = 'GH'
        assert_equal(header[1], 'GH')
        assert_equal(header[2], 'GH')

        # Now assign via an iterable
        header[1:] = ['H', 'I']
        assert_equal(header[1], 'H')
        assert_equal(header[2], 'I')

    def test_header_slice_delete(self):
        """Test deleting a slice of cards from the header."""

        header = pyfits.Header([('A', 'B'), ('C', 'D'), ('E', 'F')])
        del header[1:]
        assert_equal(len(header), 1)
        assert_equal(header[0], 'B')
        del header[:]
        assert_equal(len(header), 0)

    def test_wildcard_slice(self):
        """Test selecting a subsection of a header via wildcard matching."""

        header = pyfits.Header([('ABC', 0), ('DEF', 1), ('ABD', 2)])
        newheader = header['AB*']
        assert_equal(len(newheader), 2)
        assert_equal(newheader[0], 0)
        assert_equal(newheader[1], 2)

    def test_wildcard_slice_assignment(self):
        """Test assigning to a header slice selected via wildcard matching."""

        header = pyfits.Header([('ABC', 0), ('DEF', 1), ('ABD', 2)])

        # Test assigning slice to the same value; this works similarly to numpy
        # arrays
        header['AB*'] = 1
        assert_equal(header[0], 1)
        assert_equal(header[2], 1)

        # Though strings are iterable they should be treated as a scalar value
        header['AB*'] = 'GH'
        assert_equal(header[0], 'GH')
        assert_equal(header[2], 'GH')

        # Now assign via an iterable
        header['AB*'] = ['H', 'I']
        assert_equal(header[0], 'H')
        assert_equal(header[2], 'I')

    def test_wildcard_slice_deletion(self):
        """Test deleting cards from a header that match a wildcard pattern."""

        header = pyfits.Header([('ABC', 0), ('DEF', 1), ('ABD', 2)])
        del header['AB*']
        assert_equal(len(header), 1)
        assert_equal(header[0], 1)

    def test_header_history(self):
        header = pyfits.Header([('ABC', 0), ('HISTORY', 1), ('HISTORY', 2),
                                ('DEF', 3), ('HISTORY', 4), ('HISTORY', 5)])
        assert_equal(header['HISTORY'], [1, 2, 4, 5])

    def test_header_clear(self):
        header = pyfits.Header([('A', 'B'), ('C', 'D')])
        header.clear()
        assert_true('A' not in header)
        assert_true('C' not in header)
        assert_equal(len(header), 0)

    def test_header_fromkeys(self):
        header = pyfits.Header.fromkeys(['A', 'B'])
        assert_true('A' in header)
        assert_equal(header['A'], '')
        assert_equal(header.comments['A'], '')
        assert_true('B' in header)
        assert_equal(header['B'], '')
        assert_equal(header.comments['B'], '')

    def test_header_fromkeys_with_value(self):
        header = pyfits.Header.fromkeys(['A', 'B'], 'C')
        assert_true('A' in header)
        assert_equal(header['A'], 'C')
        assert_equal(header.comments['A'], '')
        assert_true('B' in header)
        assert_equal(header['B'], 'C')
        assert_equal(header.comments['B'], '')

    def test_header_fromkeys_with_value_and_comment(self):
        header = pyfits.Header.fromkeys(['A'], ('B', 'C'))
        assert_true('A' in header)
        assert_equal(header['A'], 'B')
        assert_equal(header.comments['A'], 'C')

    def test_header_fromkeys_with_duplicates(self):
        header = pyfits.Header.fromkeys(['A', 'B', 'A'], 'C')
        assert_true('A' in header)
        assert_true(('A', 0) in header)
        assert_true(('A', 1) in header)
        assert_true(('A', 2) not in header)
        assert_equal(header[0], 'C')
        assert_equal(header['A'], 'C')
        assert_equal(header[('A', 0)], 'C')
        assert_equal(header[2], 'C')
        assert_equal(header[('A', 1)], 'C')

    def test_header_items(self):
        header = pyfits.Header([('A', 'B'), ('C', 'D')])
        assert_equal(header.items(), list(header.iteritems()))

    def test_header_iterkeys(self):
        header = pyfits.Header([('A', 'B'), ('C', 'D')])
        for a, b in itertools.izip(header.iterkeys(), header):
            assert_equal(a, b)

    def test_header_itervalues(self):
        header = pyfits.Header([('A', 'B'), ('C', 'D')])
        for a, b in itertools.izip(header.itervalues(), ['B', 'D']):
            assert_equal(a, b)

    def test_header_keys(self):
        hdul = pyfits.open(self.data('arange.fits'))
        assert_equal(hdul[0].header.keys(),
                     ['SIMPLE', 'BITPIX', 'NAXIS', 'NAXIS1', 'NAXIS2',
                      'NAXIS3', 'EXTEND'])

    def test_header_list_like_pop(self):
        header = pyfits.Header([('A', 'B'), ('C', 'D'), ('E', 'F'),
                                ('G', 'H')])

        last = header.pop()
        assert_equal(last, 'H')
        assert_equal(len(header), 3)
        assert_equal(header.keys(), ['A', 'C', 'E'])

        mid = header.pop(1)
        assert_equal(mid, 'D')
        assert_equal(len(header), 2)
        assert_equal(header.keys(), ['A', 'E'])

        first = header.pop(0)
        assert_equal(first, 'B')
        assert_equal(len(header), 1)
        assert_equal(header.keys(), ['E'])

        assert_raises(IndexError, header.pop, 42)

    def test_header_dict_like_pop(self):
        header = pyfits.Header([('A', 'B'), ('C', 'D'), ('E', 'F'),
                                ('G', 'H')])
        assert_raises(TypeError, header.pop, 'A', 'B', 'C')

        last = header.pop('G')
        assert_equal(last, 'H')
        assert_equal(len(header), 3)
        assert_equal(header.keys(), ['A', 'C', 'E'])

        mid = header.pop('C')
        assert_equal(mid, 'D')
        assert_equal(len(header), 2)
        assert_equal(header.keys(), ['A', 'E'])

        first = header.pop('A')
        assert_equal(first, 'B')
        assert_equal(len(header), 1)
        assert_equal(header.keys(), ['E'])

        default = header.pop('X', 'Y')
        assert_equal(default, 'Y')
        assert_equal(len(header), 1)

        assert_raises(KeyError, header.pop, 'X')

    def test_popitem(self):
        header = pyfits.Header([('A', 'B'), ('C', 'D'), ('E', 'F')])
        keyword, value = header.popitem()
        assert_true(keyword not in header)
        assert_equal(len(header), 2)
        keyword, value = header.popitem()
        assert_true(keyword not in header)
        assert_equal(len(header), 1)
        keyword, value = header.popitem()
        assert_true(keyword not in header)
        assert_equal(len(header), 0)
        assert_raises(KeyError, header.popitem)

    def test_setdefault(self):
        header = pyfits.Header([('A', 'B'), ('C', 'D'), ('E', 'F')])
        assert_equal(header.setdefault('A'), 'B')
        assert_equal(header.setdefault('C'), 'D')
        assert_equal(header.setdefault('E'), 'F')
        assert_equal(len(header), 3)
        assert_equal(header.setdefault('G', 'H'), 'H')
        assert_equal(len(header), 4)
        assert_true('G' in header)
        assert_equal(header.setdefault('G', 'H'), 'H')
        assert_equal(len(header), 4)

    def test_update_from_dict(self):
        """
        Test adding new cards and updating existing cards from a dict using
        Header.update()
        """

        header = pyfits.Header([('A', 'B'), ('C', 'D')])
        header.update({'A': 'E', 'F': 'G'})
        assert_equal(header['A'], 'E')
        assert_equal(header[0], 'E')
        assert_true('F' in header)
        assert_equal(header['F'], 'G')
        assert_equal(header[-1], 'G')

        # Same as above but this time pass the update dict as keyword arguments
        header = pyfits.Header([('A', 'B'), ('C', 'D')])
        header.update(A='E', F='G')
        assert_equal(header['A'], 'E')
        assert_equal(header[0], 'E')
        assert_true('F' in header)
        assert_equal(header['F'], 'G')
        assert_equal(header[-1], 'G')

    def test_update_from_iterable(self):
        """
        Test adding new cards and updating existing cards from an iterable of
        cards and card tuples.
        """

        header = pyfits.Header([('A', 'B'), ('C', 'D')])
        header.update([('A', 'E'), pyfits.Card('F', 'G')])
        assert_equal(header['A'], 'E')
        assert_equal(header[0], 'E')
        assert_true('F' in header)
        assert_equal(header['F'], 'G')
        assert_equal(header[-1], 'G')

    def test_header_extend(self):
        """
        Test extending a header both with and without stripping cards from the
        extension header.
        """

        hdu = pyfits.PrimaryHDU()
        hdu2 = pyfits.ImageHDU()
        hdu2.header['MYKEY'] = ('some val', 'some comment')
        hdu.header += hdu2.header
        assert_equal(len(hdu.header), 5)
        assert_equal(hdu.header[-1], 'some val')

        # Same thing, but using + instead of +=
        hdu = pyfits.PrimaryHDU()
        hdu.header = hdu.header + hdu2.header
        assert_equal(len(hdu.header), 5)
        assert_equal(hdu.header[-1], 'some val')

        # Directly append the other header in full--not usually a desirable
        # operation when the header is coming from another HDU
        hdu.header.extend(hdu2.header, strip=False)
        assert_equal(len(hdu.header), 11)
        assert_equal(hdu.header.keys()[5], 'XTENSION')
        assert_equal(hdu.header[-1], 'some val')
        assert_true(('MYKEY', 1) in hdu.header)

    def test_header_extend_unique(self):
        """
        Test extending the header with and without unique=True.
        """
        hdu = pyfits.PrimaryHDU()
        hdu2 = pyfits.ImageHDU()
        hdu.header['MYKEY'] = ('some val', 'some comment')
        hdu2.header['MYKEY'] = ('some other val', 'some other comment')
        hdu.header.extend(hdu2.header)
        assert_equal(len(hdu.header), 6)
        assert_equal(hdu.header[-2], 'some val')
        assert_equal(hdu.header[-1], 'some other val')

        hdu = pyfits.PrimaryHDU()
        hdu2 = pyfits.ImageHDU()
        hdu.header['MYKEY'] = ('some val', 'some comment')
        hdu.header.extend(hdu2.header, unique=True)
        assert_equal(len(hdu.header), 5)
        assert_equal(hdu.header[-1], 'some val')

    def test_header_extend_update(self):
        """
        Test extending the header with and without update=True.
        """

        hdu = pyfits.PrimaryHDU()
        hdu2 = pyfits.ImageHDU()
        hdu.header['MYKEY'] = ('some val', 'some comment')
        hdu.header['HISTORY'] = 'history 1'
        hdu2.header['MYKEY'] = ('some other val', 'some other comment')
        hdu2.header['HISTORY'] = 'history 1'
        hdu2.header['HISTORY'] = 'history 2'
        hdu.header.extend(hdu2.header)
        assert_equal(len(hdu.header), 9)
        assert_true(('MYKEY', 0) in hdu.header)
        assert_true(('MYKEY', 1) in hdu.header)
        assert_equal(hdu.header[('MYKEY', 1)], 'some other val')
        assert_equal(len(hdu.header['HISTORY']), 3)
        assert_equal(hdu.header[-1], 'history 2')

        hdu = pyfits.PrimaryHDU()
        hdu.header['MYKEY'] = ('some val', 'some comment')
        hdu.header['HISTORY'] = 'history 1'
        hdu.header.extend(hdu2.header, update=True)
        assert_equal(len(hdu.header), 7)
        assert_true(('MYKEY', 0) in hdu.header)
        assert_false(('MYKEY', 1) in hdu.header)
        assert_equal(hdu.header['MYKEY'], 'some other val')
        assert_equal(len(hdu.header['HISTORY']), 2)
        assert_equal(hdu.header[-1], 'history 2')

    def test_header_extend_exact(self):
        """
        Test that extending an empty header with the contents of an existing
        header can exactly duplicate that header, given strip=False and
        end=True.
        """

        header = pyfits.getheader(self.data('test0.fits'))
        header2 = pyfits.Header()
        header2.extend(header, strip=False, end=True)
        assert_equal(header, header2)

    def test_header_count(self):
        header = pyfits.Header([('A', 'B'), ('C', 'D'), ('E', 'F')])
        assert_equal(header.count('A'), 1)
        assert_equal(header.count('C'), 1)
        assert_equal(header.count('E'), 1)
        header['HISTORY'] = 'a'
        header['HISTORY'] = 'b'
        assert_equal(header.count('HISTORY'), 2)
        assert_raises(KeyError, header.count, 'G')

    def test_header_append_use_blanks(self):
        """
        Tests that blank cards can be appended, and that future appends will
        use blank cards when available (unless useblanks=False)
        """

        header = pyfits.Header([('A', 'B'), ('C', 'D')])

        # Append a couple blanks
        header.append()
        header.append()
        assert_equal(len(header), 4)
        assert_equal(header[-1], '')
        assert_equal(header[-2], '')

        # New card should fill the first blank by default
        header.append(('E', 'F'))
        assert_equal(len(header), 4)
        assert_equal(header[-2], 'F')
        assert_equal(header[-1], '')

        # This card should not use up a blank spot
        header.append(('G', 'H'), useblanks=False)
        assert_equal(len(header), 5)
        assert_equal(header[-1], '')
        assert_equal(header[-2], 'H')

    def test_header_append_keyword_only(self):
        """
        Test appending a new card with just the keyword, and no value or
        comment given.
        """

        header = pyfits.Header([('A', 'B'), ('C', 'D')])
        header.append('E')
        assert_equal(len(header), 3)
        assert_equal(header.keys()[-1], 'E')
        assert_equal(header[-1], '')
        assert_equal(header.comments['E'], '')

        # Try appending a blank--normally this can be accomplished with just
        # header.append(), but header.append('') should also work (and is maybe
        # a little more clear)
        header.append('')
        assert_equal(len(header), 4)

        assert_equal(header.keys()[-1], '')
        assert_equal(header[''], '')
        assert_equal(header.comments[''], '')

    def test_header_insert_use_blanks(self):
        header = pyfits.Header([('A', 'B'), ('C', 'D')])

        # Append a couple blanks
        header.append()
        header.append()

        # Insert a new card; should use up one of the blanks
        header.insert(1, ('E', 'F'))
        assert_equal(len(header), 4)
        assert_equal(header[1], 'F')
        assert_equal(header[-1], '')
        assert_equal(header[-2], 'D')

        # Insert a new card without using blanks
        header.insert(1, ('G', 'H'), useblanks=False)
        assert_equal(len(header), 5)
        assert_equal(header[1], 'H')
        assert_equal(header[-1], '')

    def test_header_comments(self):
        header = pyfits.Header([('A', 'B', 'C'), ('DEF', 'G', 'H')])
        assert_equal(repr(header.comments),
                     '       A  C\n'
                     '     DEF  H')

    def test_comment_slices_and_filters(self):
        header = pyfits.Header([('AB', 'C', 'D'), ('EF', 'G', 'H'),
                                ('AI', 'J', 'K')])
        s = header.comments[1:]
        assert_equal(list(s), ['H', 'K'])
        s = header.comments[::-1]
        assert_equal(list(s), ['K', 'H', 'D'])
        s = header.comments['A*']
        assert_equal(list(s), ['D', 'K'])

    def test_comment_slice_filter_assign(self):
        header = pyfits.Header([('AB', 'C', 'D'), ('EF', 'G', 'H'),
                                ('AI', 'J', 'K')])
        header.comments[1:] = 'L'
        assert_equal(list(header.comments), ['D', 'L', 'L'])
        assert_equal(header.cards[header.index('AB')].comment, 'D')
        assert_equal(header.cards[header.index('EF')].comment, 'L')
        assert_equal(header.cards[header.index('AI')].comment, 'L')

        header.comments[::-1] = header.comments[:]
        assert_equal(list(header.comments), ['L', 'L', 'D'])

        header.comments['A*'] = ['M', 'N']
        assert_equal(list(header.comments), ['M', 'L', 'N'])

    def test_update_comment(self):
        hdul = pyfits.open(self.data('arange.fits'))
        hdul[0].header['FOO'] = ('BAR', 'BAZ')
        hdul.writeto(self.temp('test.fits'))

        hdul = pyfits.open(self.temp('test.fits'), mode='update')
        hdul[0].header.comments['FOO'] = 'QUX'
        hdul.close()

        hdul = pyfits.open(self.temp('test.fits'))
        assert_equal(hdul[0].header.comments['FOO'], 'QUX')

    def test_commentary_slicing(self):
        header = pyfits.Header()

        indices = range(5)

        for idx in indices:
            header['HISTORY'] = idx

        # Just a few sample slice types; this won't get all corner cases but if
        # these all work we should be in good shape
        assert_equal(header['HISTORY'][1:], indices[1:])
        assert_equal(header['HISTORY'][:3], indices[:3])
        assert_equal(header['HISTORY'][:6], indices[:6])
        assert_equal(header['HISTORY'][:-2], indices[:-2])
        assert_equal(header['HISTORY'][::-1], indices[::-1])
        assert_equal(header['HISTORY'][1::-1], indices[1::-1])
        assert_equal(header['HISTORY'][1:5:2], indices[1:5:2])

        # Same tests, but copy the values first; as it turns out this is
        # different from just directly doing an __eq__ as in the first set of
        # assertions
        header.insert(0, ('A', 'B', 'C'))
        header.append(('D', 'E', 'F'), end=True)
        assert_equal(list(header['HISTORY'][1:]), indices[1:])
        assert_equal(list(header['HISTORY'][:3]), indices[:3])
        assert_equal(list(header['HISTORY'][:6]), indices[:6])
        assert_equal(list(header['HISTORY'][:-2]), indices[:-2])
        assert_equal(list(header['HISTORY'][::-1]), indices[::-1])
        assert_equal(list(header['HISTORY'][1::-1]), indices[1::-1])
        assert_equal(list(header['HISTORY'][1:5:2]), indices[1:5:2])

    def test_update_commentary(self):
        header = pyfits.Header()
        header['FOO'] = 'BAR'
        header['HISTORY'] = 'ABC'
        header['FRED'] = 'BARNEY'
        header['HISTORY'] = 'DEF'
        header['HISTORY'] = 'GHI'

        assert_equal(header['HISTORY'], ['ABC', 'DEF', 'GHI'])

        # Single value update
        header['HISTORY'][0] = 'FOO'
        assert_equal(header['HISTORY'], ['FOO', 'DEF', 'GHI'])

        # Single value partial slice update
        header['HISTORY'][1:] = 'BAR'
        assert_equal(header['HISTORY'], ['FOO', 'BAR', 'BAR'])

        # Multi-value update
        header['HISTORY'][:] = ['BAZ', 'QUX']
        assert_equal(header['HISTORY'], ['BAZ', 'QUX', 'BAR'])

    def test_long_commentary_card(self):
        header = pyfits.Header()
        header['FOO'] = 'BAR'
        header['BAZ'] = 'QUX'
        longval = 'ABC' * 30
        header['HISTORY'] = longval
        header['FRED'] = 'BARNEY'
        header['HISTORY'] = longval

        assert_equal(len(header), 7)
        assert_equal(header.keys()[2], 'FRED')
        assert_equal(str(header.cards[3]), 'HISTORY ' + longval[:72])
        assert_equal(str(header.cards[4]).rstrip(), 'HISTORY ' + longval[72:])

        header.set('HISTORY', longval, after='FOO')
        assert_equal(len(header), 9)
        assert_equal(str(header.cards[1]), 'HISTORY ' + longval[:72])
        assert_equal(str(header.cards[2]).rstrip(), 'HISTORY ' + longval[72:])

    def test_header_fromtextfile(self):
        """Regression test for #122.

        Manually write a text file containing some header cards ending with
        newlines and ensure that fromtextfile can read them back in.
        """

        header = pyfits.Header()
        header['A'] = ('B', 'C')
        header['B'] = ('C', 'D')
        header['C'] = ('D', 'E')

        with open(self.temp('test.hdr'), 'w') as f:
            f.write('\n'.join(str(c).strip() for c in header.cards))

        header2 = pyfits.Header.fromtextfile(self.temp('test.hdr'))
        assert_equal(header, header2)

    def test_header_fromtextfile_with_end_card(self):
        """Regression test for #154.

        Make sure that when a Header is read from a text file that the END card
        is ignored.
        """

        header = pyfits.Header([('A', 'B', 'C'), ('D', 'E', 'F')])

        # We don't use header.totextfile here because it writes each card with
        # trailing spaces to pad them out to 80 characters.  But this bug only
        # presents itself when each card ends immediately with a newline, and
        # no trailing spaces
        with open(self.temp('test.hdr'), 'w') as f:
            f.write('\n'.join(str(c).strip() for c in header.cards))
            f.write('\nEND')

        new_header = pyfits.Header.fromtextfile(self.temp('test.hdr'))

        assert_false('END' in new_header)
        assert_equal(header, new_header)

    def test_append_end_card(self):
        """
        Regression test 2 for #154.

        Manually adding an END card to a header should simply result in a
        ValueError (as was the case in PyFITS 3.0 and earlier).
        """

        header = pyfits.Header([('A', 'B', 'C'), ('D', 'E', 'F')])

        def setitem(k, v):
            header[k] = v

        assert_raises(ValueError, setitem, 'END', '')
        assert_raises(ValueError, header.append, 'END')
        assert_raises(ValueError, header.append, 'END', end=True)
        assert_raises(ValueError, header.insert, len(header), 'END')
        assert_raises(ValueError, header.set, 'END')

    def test_unnecessary_move(self):
        """
        Regression test for #125.

        Ensures that a header is not modified when setting the position of a
        keyword that's already in its correct position.
        """

        header = pyfits.Header([('A', 'B'), ('B', 'C'), ('C', 'D')])

        header.set('B', before=2)
        assert_equal(header.keys(), ['A', 'B', 'C'])
        assert_false(header._modified)

        header.set('B', after=0)
        assert_equal(header.keys(), ['A', 'B', 'C'])
        assert_false(header._modified)

        header.set('B', before='C')
        assert_equal(header.keys(), ['A', 'B', 'C'])
        assert_false(header._modified)

        header.set('B', after='A')
        assert_equal(header.keys(), ['A', 'B', 'C'])
        assert_false(header._modified)

        header.set('B', before=2)
        assert_equal(header.keys(), ['A', 'B', 'C'])
        assert_false(header._modified)

        # 123 is well past the end, and C is already at the end, so it's in the
        # right place already
        header.set('C', before=123)
        assert_equal(header.keys(), ['A', 'B', 'C'])
        assert_false(header._modified)

        header.set('C', after=123)
        assert_equal(header.keys(), ['A', 'B', 'C'])
        assert_false(header._modified)

    def test_invalid_float_cards(self):
        """Regression test for #137."""

        # Create a header containing two of the problematic cards in the test
        # case where this came up:
        hstr = "FOCALLEN= +1.550000000000e+002\nAPERTURE= +0.000000000000e+000"
        h = pyfits.Header.fromstring(hstr, sep='\n')

        # First the case that *does* work prior to fixing this issue
        assert_equal(h['FOCALLEN'], 155.0)
        assert_equal(h['APERTURE'], 0.0)

        # Now if this were reserialized, would new values for these cards be
        # written with repaired exponent signs?
        with CaptureStdio():
            assert_equal(str(h.cards['FOCALLEN']),
                         _pad("FOCALLEN= +1.550000000000E+002"))
            assert_true(h.cards['FOCALLEN']._modified)
            assert_equal(str(h.cards['APERTURE']),
                         _pad("APERTURE= +0.000000000000E+000"))
            assert_true(h.cards['APERTURE']._modified)
            assert_true(h._modified)

        # This is the case that was specifically causing problems; generating
        # the card strings *before* parsing the values.  Also, the card strings
        # really should be "fixed" before being returned to the user
        h = pyfits.Header.fromstring(hstr, sep='\n')
        with CaptureStdio():
            assert_equal(str(h.cards['FOCALLEN']),
                         _pad("FOCALLEN= +1.550000000000E+002"))
            assert_true(h.cards['FOCALLEN']._modified)
            assert_equal(str(h.cards['APERTURE']),
                         _pad("APERTURE= +0.000000000000E+000"))
            assert_true(h.cards['APERTURE']._modified)

        assert_equal(h['FOCALLEN'], 155.0)
        assert_equal(h['APERTURE'], 0.0)
        assert_true(h._modified)

        # For the heck of it, try assigning the identical values and ensure
        # that the newly fixed value strings are left intact
        h['FOCALLEN'] = 155.0
        h['APERTURE'] = 0.0
        assert_equal(str(h.cards['FOCALLEN']),
                     _pad("FOCALLEN= +1.550000000000E+002"))
        assert_equal(str(h.cards['APERTURE']),
                     _pad("APERTURE= +0.000000000000E+000"))

    def test_invalid_float_cards2(self):
        """
        Regression test for #140.
        """

        # The example for this test requires creating a FITS file containing a
        # slightly misformatted float value.  I can't actually even find a way
        # to do that directly through PyFITS--it won't let me.
        hdu = pyfits.PrimaryHDU()
        hdu.header['TEST'] = 5.0022221e-07
        hdu.writeto(self.temp('test.fits'))

        # Here we manually make the file invalid
        with open(self.temp('test.fits'), 'rb+') as f:
            f.seek(346)  # Location of the exponent 'E' symbol
            f.write(encode_ascii('e'))

        hdul = pyfits.open(self.temp('test.fits'))
        with catch_warnings(record=True) as w:
            with CaptureStdio():
                hdul.writeto(self.temp('temp.fits'), output_verify='warn')
            assert_equal(len(w), 5)
            # The first two warnings are just the headers to the actual warning
            # message (HDU 0, Card 4).  I'm still not sure things like that
            # should be output as separate warning messages, but that's
            # something to think about...
            msg = str(w[3].message)
            assert_true('(invalid value string: 5.0022221e-07)' in msg)

    def test_leading_zeros(self):
        """
        Regression test for #137, part 2.

        Ticket #137 also showed that in float values like 0.001 the leading
        zero was unnecessarily being stripped off when rewriting the header.
        Though leading zeros should be removed from integer values to prevent
        misinterpretation as octal by python (for now PyFITS will still
        maintain the leading zeros if now changes are made to the value, but
        will drop them if changes are made).
        """

        c = pyfits.Card.fromstring("APERTURE= +0.000000000000E+000")
        assert_equal(str(c), _pad("APERTURE= +0.000000000000E+000"))
        assert_equal(c.value, 0.0)
        c = pyfits.Card.fromstring("APERTURE= 0.000000000000E+000")
        assert_equal(str(c), _pad("APERTURE= 0.000000000000E+000"))
        assert_equal(c.value, 0.0)
        c = pyfits.Card.fromstring("APERTURE= 017")
        assert_equal(str(c), _pad("APERTURE= 017"))
        assert_equal(c.value, 17)

    def test_assign_boolean(self):
        """
        Regression test for #123. Tests assigning Python and Numpy boolean
        values to keyword values.
        """

        fooimg = _pad('FOO     =                    T')
        barimg = _pad('BAR     =                    F')
        h = pyfits.Header()
        h['FOO'] = True
        h['BAR'] = False
        assert_equal(h['FOO'], True)
        assert_equal(h['BAR'], False)
        assert_equal(str(h.cards['FOO']), fooimg)
        assert_equal(str(h.cards['BAR']), barimg)

        h = pyfits.Header()
        h['FOO'] = np.bool_(True)
        h['BAR'] = np.bool_(False)
        assert_equal(h['FOO'], True)
        assert_equal(h['BAR'], False)
        assert_equal(str(h.cards['FOO']), fooimg)
        assert_equal(str(h.cards['BAR']), barimg)

        h = pyfits.Header()
        h.append(pyfits.Card.fromstring(fooimg))
        h.append(pyfits.Card.fromstring(barimg))
        assert_equal(h['FOO'], True)
        assert_equal(h['BAR'], False)
        assert_equal(str(h.cards['FOO']), fooimg)
        assert_equal(str(h.cards['BAR']), barimg)

    def test_header_method_keyword_normalization(self):
        """
        Regression test for #149.  Basically ensures that all public Header
        methods are case-insensitive w.r.t. keywords.

        Provides a reasonably comprehensive test of several methods at once.
        """

        h = pyfits.Header([('abC', 1), ('Def', 2), ('GeH', 3)])
        assert_equal(h.keys(), ['ABC', 'DEF', 'GEH'])
        assert_true('abc' in h)
        assert_true('dEf' in h)

        assert_equal(h['geh'], 3)

        # Case insensitivity of wildcards
        assert_equal(len(h['g*']), 1)

        h['aBc'] = 2
        assert_equal(h['abc'], 2)
        # ABC already existed so assigning to aBc should not have added any new
        # cards
        assert_equal(len(h), 3)

        del h['gEh']
        assert_equal(h.keys(), ['ABC', 'DEF'])
        assert_equal(len(h), 2)
        assert_equal(h.get('def'), 2)

        h.set('Abc', 3)
        assert_equal(h['ABC'], 3)
        h.set('gEh', 3, before='Abc')
        assert_equal(h.keys(), ['GEH', 'ABC', 'DEF'])

        assert_equal(h.pop('abC'), 3)
        assert_equal(len(h), 2)

        assert_equal(h.setdefault('def', 3), 2)
        assert_equal(len(h), 2)
        assert_equal(h.setdefault('aBc', 1), 1)
        assert_equal(len(h), 3)
        assert_equal(h.keys(), ['GEH', 'DEF', 'ABC'])

        h.update({'GeH': 1, 'iJk': 4})
        assert_equal(len(h), 4)
        assert_equal(h.keys(), ['GEH', 'DEF', 'ABC', 'IJK'])
        assert_equal(h['GEH'], 1)

        assert_equal(h.count('ijk'), 1)
        assert_equal(h.index('ijk'), 3)

        h.remove('Def')
        assert_equal(len(h), 3)
        assert_equal(h.keys(), ['GEH', 'ABC', 'IJK'])

    def test_end_in_comment(self):
        """
        Regression test for #142.  Tests a case where the comment of a card
        ends with END, and is followed by several blank cards.
        """

        data = np.arange(100).reshape((10, 10))
        hdu = pyfits.PrimaryHDU(data=data)
        hdu.header['TESTKW'] = ('Test val', 'This is the END')
        # Add a couple blanks after the END string
        hdu.header.append()
        hdu.header.append()
        hdu.writeto(self.temp('test.fits'))

        with pyfits.open(self.temp('test.fits'), memmap=False) as hdul:
            # memmap = False to avoid leaving open a mmap to the file when we
            # access the data--this causes problems on Windows when we try to
            # overwrite the file later
            assert_true('TESTKW' in hdul[0].header)
            assert_equal(hdul[0].header, hdu.header)
            assert_true((hdul[0].data == data).all())

        # Add blanks until the header is extended to two block sizes
        while len(hdu.header) < 36:
            hdu.header.append()
        with ignore_warnings():
            hdu.writeto(self.temp('test.fits'), clobber=True)

        with pyfits.open(self.temp('test.fits')) as hdul:
            assert_true('TESTKW' in hdul[0].header)
            assert_equal(hdul[0].header, hdu.header)
            assert_true((hdul[0].data == data).all())

        # Test parsing the same header when it's written to a text file
        hdu.header.totextfile(self.temp('test.hdr'))
        header2 = pyfits.Header.fromtextfile(self.temp('test.hdr'))
        assert_equal(hdu.header, header2)

    def test_assign_unicode(self):
        """
        Regression test for #134.  Assigning a unicode literal as a header
        value should not fail silently.  If the value can be converted to ASCII
        then it should just work.  Otherwise it should fail with an appropriate
        value error.

        Also tests unicode for keywords and comments.
        """

        erikku = u'\u30a8\u30ea\u30c3\u30af'

        def assign(keyword, val):
            h[keyword] = val

        h = pyfits.Header()
        h[u'FOO'] = 'BAR'
        assert_true('FOO' in h)
        assert_equal(h['FOO'], 'BAR')
        assert_equal(h[u'FOO'], 'BAR')
        assert_equal(repr(h), _pad("FOO     = 'BAR     '"))
        assert_raises(ValueError, assign, erikku, 'BAR')


        h['FOO'] = u'BAZ'
        assert_equal(h[u'FOO'], 'BAZ')
        assert_equal(h[u'FOO'], u'BAZ')
        assert_equal(repr(h), _pad("FOO     = 'BAZ     '"))
        assert_raises(ValueError, assign, 'FOO', erikku)

        h['FOO'] = ('BAR', u'BAZ')
        assert_equal(h['FOO'], 'BAR')
        assert_equal(h['FOO'], u'BAR')
        assert_equal(h.comments['FOO'], 'BAZ')
        assert_equal(h.comments['FOO'], u'BAZ')
        assert_equal(repr(h), _pad("FOO     = 'BAR     '           / BAZ"))

        h['FOO'] = (u'BAR', u'BAZ')
        assert_equal(h['FOO'], 'BAR')
        assert_equal(h['FOO'], u'BAR')
        assert_equal(h.comments['FOO'], 'BAZ')
        assert_equal(h.comments['FOO'], u'BAZ')
        assert_equal(repr(h), _pad("FOO     = 'BAR     '           / BAZ"))

        assert_raises(ValueError, assign, 'FOO', ('BAR', erikku))
        assert_raises(ValueError, assign, 'FOO', (erikku, 'BAZ'))
        assert_raises(ValueError, assign, 'FOO', (erikku, erikku))

    def test_header_strip_whitespace(self):
        """
        Regression test for #146, and for the solution that is optional
        stripping of whitespace from the end of a header value.

        By default extra whitespace is stripped off, but if
        pyfits.STRIP_HEADER_WHITESPACE = False it should not be stripped.
        """

        h = pyfits.Header()
        h['FOO'] = 'Bar      '
        assert_equal(h['FOO'], 'Bar')
        c = pyfits.Card.fromstring("QUX     = 'Bar        '")
        h.append(c)
        assert_equal(h['QUX'], 'Bar')
        assert_equal(h.cards['FOO'].image.rstrip(),
                     "FOO     = 'Bar      '")
        assert_equal(h.cards['QUX'].image.rstrip(),
                     "QUX     = 'Bar        '")

        pyfits.STRIP_HEADER_WHITESPACE = False
        try:
            assert_equal(h['FOO'], 'Bar      ')
            assert_equal(h['QUX'], 'Bar        ')
            assert_equal(h.cards['FOO'].image.rstrip(),
                         "FOO     = 'Bar      '")
            assert_equal(h.cards['QUX'].image.rstrip(),
                         "QUX     = 'Bar        '")
        finally:
            pyfits.STRIP_HEADER_WHITESPACE = True

        assert_equal(h['FOO'], 'Bar')
        assert_equal(h['QUX'], 'Bar')
        assert_equal(h.cards['FOO'].image.rstrip(),
                     "FOO     = 'Bar      '")
        assert_equal(h.cards['QUX'].image.rstrip(),
                     "QUX     = 'Bar        '")

    def test_keep_duplicate_history_in_orig_header(self):
        """
        Regression test for #156.

        When creating a new HDU from an existing Header read from an existing
        FITS file, if the origianl header contains duplicate HISTORY values
        those duplicates should be preserved just as in the original header.

        This bug occurred due to naivete in Header.extend.
        """

        history = ['CCD parameters table ...',
                   '   reference table oref$n951041ko_ccd.fits',
                   '     INFLIGHT 12/07/2001 25/02/2002',
                   '     all bias frames'] * 3

        hdu = pyfits.PrimaryHDU()
        # Add the history entries twice
        for item in history:
            hdu.header['HISTORY'] = item

        hdu.writeto(self.temp('test.fits'))

        with pyfits.open(self.temp('test.fits')) as hdul:
            assert_equal(hdul[0].header['HISTORY'], history)

        new_hdu = pyfits.PrimaryHDU(header=hdu.header)
        assert_equal(new_hdu.header['HISTORY'], hdu.header['HISTORY'])
        new_hdu.writeto(self.temp('test2.fits'))

        with pyfits.open(self.temp('test2.fits')) as hdul:
            assert_equal(hdul[0].header['HISTORY'], history)

    def test_invalid_keyword_cards(self):
        """
        Test for #109.  Allow opening files with headers containing invalid
        keywords.
        """

        # Create a header containing a few different types of BAD headers.
        c1 = pyfits.Card.fromstring('CLFIND2D: contour = 0.30')
        c2 = pyfits.Card.fromstring('Just some random text.')
        c3 = pyfits.Card.fromstring('A' * 80)

        hdu = pyfits.PrimaryHDU()
        # This should work with some warnings
        with catch_warnings(record=True) as w:
            hdu.header.append(c1)
            hdu.header.append(c2)
            hdu.header.append(c3)
            assert_equal(len(w), 3)

        hdu.writeto(self.temp('test.fits'))

        with catch_warnings(record=True) as w:
            with pyfits.open(self.temp('test.fits')) as hdul:
                # Merely opening the file should blast some warnings about the
                # invalid keywords
                assert_equal(len(w), 3)

                header = hdul[0].header
                assert_true('CLFIND2D' in header)
                assert_true('Just som' in header)
                assert_true('AAAAAAAA' in header)

                assert_equal(header['CLFIND2D'], ': contour = 0.30')
                assert_equal(header['Just som'], 'e random text.')
                assert_equal(header['AAAAAAAA'], 'A' * 72)

                # It should not be possible to assign to the invalid keywords
                assert_raises(ValueError, header.set, 'CLFIND2D', 'foo')
                assert_raises(ValueError, header.set, 'Just som', 'foo')
                assert_raises(ValueError, header.set, 'AAAAAAAA', 'foo')

    def test_fix_hierarch_with_invalid_value(self):
        """
        Regression test for #172.  Ensures that when fixing a hierarch card it
        remains a hierarch card.
        """

        c = pyfits.Card.fromstring('HIERARCH ESO DET CHIP PXSPACE = 5e6')
        with CaptureStdio():
            c.verify('fix')
        assert_equal(str(c), _pad('HIERARCH ESO DET CHIP PXSPACE = 5E6'))


class TestRecordValuedKeywordCards(PyfitsTestCase):
    """
    Tests for handling of record-valued keyword cards as used by the FITS WCS
    Paper IV proposal.

    These tests are derived primarily from the release notes for PyFITS 1.4 (in
    which this feature was first introduced.
    """

    def setup(self):
        super(TestRecordValuedKeywordCards, self).setup()
        self._test_header = pyfits.Header()
        self._test_header.set('DP1', 'NAXIS: 2')
        self._test_header.set('DP1', 'AXIS.1: 1')
        self._test_header.set('DP1', 'AXIS.2: 2')
        self._test_header.set('DP1', 'NAUX: 2')
        self._test_header.set('DP1', 'AUX.1.COEFF.0: 0')
        self._test_header.set('DP1', 'AUX.1.POWER.0: 1')
        self._test_header.set('DP1', 'AUX.1.COEFF.1: 0.00048828125')
        self._test_header.set('DP1', 'AUX.1.POWER.1: 1')

    def test_initialize_rvkc(self):
        """
        Test different methods for initializing a card that should be
        recognized as a RVKC
        """

        c = pyfits.Card.fromstring("DP1     = 'NAXIS: 2' / A comment")
        assert_equal(c.keyword, 'DP1.NAXIS')
        assert_equal(c.value, 2.0)
        assert_equal(c.field_specifier, 'NAXIS')
        assert_equal(c.comment, 'A comment')

        c = pyfits.Card.fromstring("DP1     = 'NAXIS: 2.1'")
        assert_equal(c.keyword, 'DP1.NAXIS')
        assert_equal(c.value, 2.1)
        assert_equal(c.field_specifier, 'NAXIS')

        c = pyfits.Card.fromstring("DP1     = 'NAXIS: a'")
        assert_equal(c.keyword, 'DP1')
        assert_equal(c.value, 'NAXIS: a')
        assert_equal(c.field_specifier, None)

        c = pyfits.Card('DP1', 'NAXIS: 2')
        assert_equal(c.keyword, 'DP1.NAXIS')
        assert_equal(c.value, 2.0)
        assert_equal(c.field_specifier, 'NAXIS')

        c = pyfits.Card('DP1', 'NAXIS: 2.0')
        assert_equal(c.keyword, 'DP1.NAXIS')
        assert_equal(c.value, 2.0)
        assert_equal(c.field_specifier, 'NAXIS')

        c = pyfits.Card('DP1', 'NAXIS: a')
        assert_equal(c.keyword, 'DP1')
        assert_equal(c.value, 'NAXIS: a')
        assert_equal(c.field_specifier, None)

        c = pyfits.Card('DP1.NAXIS', 2)
        assert_equal(c.keyword, 'DP1.NAXIS')
        assert_equal(c.value, 2.0)
        assert_equal(c.field_specifier, 'NAXIS')

        c = pyfits.Card('DP1.NAXIS', 2.0)
        assert_equal(c.keyword, 'DP1.NAXIS')
        assert_equal(c.value, 2.0)
        assert_equal(c.field_specifier, 'NAXIS')

        with ignore_warnings():
            c = pyfits.Card('DP1.NAXIS', 'a')
        assert_equal(c.keyword, 'DP1.NAXIS')
        assert_equal(c.value, 'a')
        assert_equal(c.field_specifier, None)

    def test_parse_field_specifier(self):
        """
        Tests that the field_specifier can accessed from a card read from a
        string before any other attributes are accessed.
        """

        c = pyfits.Card.fromstring("DP1     = 'NAXIS: 2' / A comment")
        assert_equal(c.field_specifier, 'NAXIS')
        assert_equal(c.keyword, 'DP1.NAXIS')
        assert_equal(c.value, 2.0)
        assert_equal(c.comment, 'A comment')

    def test_update_field_specifier(self):
        """
        Test setting the field_specifier attribute and updating the card image
        to reflect the new value.
        """

        c = pyfits.Card.fromstring("DP1     = 'NAXIS: 2' / A comment")
        assert_equal(c.field_specifier, 'NAXIS')
        c.field_specifier = 'NAXIS1'
        assert_equal(c.field_specifier, 'NAXIS1')
        assert_equal(c.keyword, 'DP1.NAXIS1')
        assert_equal(c.value, 2.0)
        assert_equal(c.comment, 'A comment')
        assert_equal(str(c).rstrip(), "DP1     = 'NAXIS1: 2' / A comment")

    def test_field_specifier_case_senstivity(self):
        """
        The keyword portion of an RVKC should still be case-insensitive, but
        the field-specifier portion should be case-sensitive.
        """

        header = pyfits.Header()
        header.set('abc.def', 1)
        header.set('abc.DEF', 2)
        assert_equal(header['abc.def'], 1)
        assert_equal(header['ABC.def'], 1)
        assert_equal(header['aBc.def'], 1)
        assert_equal(header['ABC.DEF'], 2)
        assert_false('ABC.dEf' in header)

    def test_get_rvkc_by_index(self):
        """
        Returning a RVKC from a header via index lookup should return the
        float value of the card.
        """

        assert_equal(self._test_header[0], 2.0)
        assert_true(isinstance(self._test_header[0], float))
        assert_equal(self._test_header[1], 1.0)
        assert_true(isinstance(self._test_header[1], float))

    def test_get_rvkc_by_keyword(self):
        """
        Returning a RVKC just via the keyword name should return the full value
        string of the first card with that keyword.

        This test was changed to reflect the requirement in ticket
        #184--previously it required _test_header['DP1'] to return the parsed
        float value.
        """

        assert_equal(self._test_header['DP1'], 'NAXIS: 2')

    def test_get_rvkc_by_keyword_and_field_specifier(self):
        """
        Returning a RVKC via the full keyword/field-specifier combination
        should return the floating point value associated with the RVKC.
        """

        assert_equal(self._test_header['DP1.NAXIS'], 2.0)
        assert_true(isinstance(self._test_header['DP1.NAXIS'], float))
        assert_equal(self._test_header['DP1.AUX.1.COEFF.1'], 0.00048828125)

    def test_access_nonexistent_rvkc(self):
        """
        Accessing a nonexistent RVKC should raise an IndexError for
        index-based lookup, or a KeyError for keyword lookup (like a normal
        card).
        """

        assert_raises(IndexError, lambda x: self._test_header[x], 8)
        assert_raises(KeyError, lambda k: self._test_header[k], 'DP1.AXIS.3')
        # Test the exception message
        try:
            self._test_header['DP1.AXIS.3']
        except KeyError, e:
            assert_equal(e.args[0], "Keyword 'DP1.AXIS.3' not found.")

    def test_update_rvkc(self):
        """A RVKC can be updated either via index or keyword access."""

        self._test_header[0] = 3
        assert_equal(self._test_header['DP1.NAXIS'], 3.0)
        assert_true(isinstance(self._test_header['DP1.NAXIS'], float))

        self._test_header['DP1.AXIS.1'] = 1.1
        assert_equal(self._test_header['DP1.AXIS.1'], 1.1)

    def test_rvkc_insert_after(self):
        """
        It should be possible to insert a new RVKC after an existing one
        specified by the full keyword/field-specifier combination."""

        self._test_header.set('DP1', 'AXIS.3: 1', 'a comment',
                              after='DP1.AXIS.2')
        assert_equal(self._test_header[3], 1)
        assert_equal(self._test_header['DP1.AXIS.3'], 1)

    def test_rvkc_delete(self):
        """
        Deleting a RVKC should work as with a normal card by using the full
        keyword/field-spcifier combination.
        """

        del self._test_header['DP1.AXIS.1']
        assert_equal(len(self._test_header), 7)
        assert_equal(self._test_header.keys()[0], 'DP1.NAXIS')
        assert_equal(self._test_header[0], 2)
        assert_equal(self._test_header.keys()[1], 'DP1.AXIS.2')
        assert_equal(self._test_header[1], 2)

    def test_pattern_matching_keys(self):
        """Test the keyword filter strings with RVKCs."""

        cl = self._test_header['DP1.AXIS.*']
        assert_true(isinstance(cl, pyfits.Header))
        assert_equal(
            [str(c).strip() for c in cl.cards],
            ["DP1     = 'AXIS.1: 1'",
             "DP1     = 'AXIS.2: 2'"])

        cl = self._test_header['DP1.N*']
        assert_equal(
            [str(c).strip() for c in cl.cards],
            ["DP1     = 'NAXIS: 2'",
             "DP1     = 'NAUX: 2'"])

        cl = self._test_header['DP1.AUX...']
        assert_equal(
            [str(c).strip() for c in cl.cards],
            ["DP1     = 'AUX.1.COEFF.0: 0'",
             "DP1     = 'AUX.1.POWER.0: 1'",
             "DP1     = 'AUX.1.COEFF.1: 0.00048828125'",
             "DP1     = 'AUX.1.POWER.1: 1'"])

        cl = self._test_header['DP?.NAXIS']
        assert_equal(
            [str(c).strip() for c in cl.cards],
            ["DP1     = 'NAXIS: 2'"])

        cl = self._test_header['DP1.A*S.*']
        assert_equal(
            [str(c).strip() for c in cl.cards],
            ["DP1     = 'AXIS.1: 1'",
             "DP1     = 'AXIS.2: 2'"])

    def test_pattern_matching_key_deletion(self):
        """Deletion by filter strings should work."""

        del self._test_header['DP1.A*...']
        assert_equal(len(self._test_header), 2)
        assert_equal(self._test_header.keys()[0], 'DP1.NAXIS')
        assert_equal(self._test_header[0], 2)
        assert_equal(self._test_header.keys()[1], 'DP1.NAUX')
        assert_equal(self._test_header[1], 2)

    def test_successive_pattern_matching(self):
        """
        A card list returned via a filter string should be further filterable.
        """

        cl = self._test_header['DP1.A*...']
        assert_equal(
            [str(c).strip() for c in cl.cards],
            ["DP1     = 'AXIS.1: 1'",
             "DP1     = 'AXIS.2: 2'",
             "DP1     = 'AUX.1.COEFF.0: 0'",
             "DP1     = 'AUX.1.POWER.0: 1'",
             "DP1     = 'AUX.1.COEFF.1: 0.00048828125'",
             "DP1     = 'AUX.1.POWER.1: 1'"])

        cl2 = cl['*.*AUX...']
        assert_equal(
            [str(c).strip() for c in cl2.cards],
            ["DP1     = 'AUX.1.COEFF.0: 0'",
             "DP1     = 'AUX.1.POWER.0: 1'",
             "DP1     = 'AUX.1.COEFF.1: 0.00048828125'",
             "DP1     = 'AUX.1.POWER.1: 1'"])

    def test_rvkc_in_cardlist_keys(self):
        """
        The CardList.keys() method should return full keyword/field-spec values
        for RVKCs.
        """

        cl = self._test_header['DP1.AXIS.*']
        assert_equal(cl.keys(), ['DP1.AXIS.1', 'DP1.AXIS.2'])

    def test_rvkc_in_cardlist_values(self):
        """
        The CardList.values() method should return the values of all RVKCs as
        floating point values.
        """

        cl = self._test_header['DP1.AXIS.*']
        assert_equal(cl.values(), [1.0, 2.0])

    def test_rvkc_value_attribute(self):
        """
        Individual card values should be accessible by the .value attribute
        (which should return a float).
        """

        cl = self._test_header['DP1.AXIS.*']
        assert_equal(cl.cards[0].value, 1.0)
        assert_true(isinstance(cl.cards[0].value, float))

    def test_overly_permissive_parsing(self):
        """
        Regression test for #183.

        Ensures that cards with standard commentary keywords are never treated
        as RVKCs.  Also ensures that cards not stricly matching the RVKC
        pattern are not treated as such.
        """

        h = pyfits.Header()
        h['HISTORY'] = 'AXIS.1: 2'
        h['HISTORY'] = 'AXIS.2: 2'
        assert_false('HISTORY.AXIS' in h)
        assert_false('HISTORY.AXIS.1' in h)
        assert_false('HISTORY.AXIS.2' in h)
        assert_equal(h['HISTORY'], ['AXIS.1: 2', 'AXIS.2: 2'])

        # This is an example straight out of the ticket where everything after
        # the '2012' in the date value was being ignored, allowing the value to
        # successfully be parsed as a "float"
        h = pyfits.Header()
        h['HISTORY'] = 'Date: 2012-09-19T13:58:53.756061'
        assert_false('HISTORY.Date' in h)
        assert_equal(str(h.cards[0]),
                     _pad('HISTORY Date: 2012-09-19T13:58:53.756061'))

        c = pyfits.Card.fromstring(
            "        'Date: 2012-09-19T13:58:53.756061'")
        assert_equal(c.keyword, '')
        assert_equal(c.value, "'Date: 2012-09-19T13:58:53.756061'")
        assert_equal(c.field_specifier, None)

        h = pyfits.Header()
        h['FOO'] = 'Date: 2012-09-19T13:58:53.756061'
        assert_false('FOO.Date' in h)
        assert_equal(str(h.cards[0]),
                     _pad("FOO     = 'Date: 2012-09-19T13:58:53.756061'"))

    def test_overly_aggressive_rvkc_lookup(self):
        """
        Regression test for #184.

        Ensures that looking up a RVKC by keyword only (without the
        field-specifier) in a header returns the full string value of that card
        without parsing it as a RVKC.  Also ensures that a full field-specifier
        is required to match a RVKC--a partial field-specifier that doesn't
        explicitly match any record-valued keyword should result in a KeyError.
        """

        c1 = pyfits.Card.fromstring("FOO     = 'AXIS.1: 2'")
        c2 = pyfits.Card.fromstring("FOO     = 'AXIS.2: 4'")
        h = pyfits.Header([c1, c2])
        assert_equal(h['FOO'], 'AXIS.1: 2')
        assert_equal(h[('FOO', 1)], 'AXIS.2: 4')
        assert_equal(h['FOO.AXIS.1'], 2.0)
        assert_equal(h['FOO.AXIS.2'], 4.0)
        assert_false('FOO.AXIS' in h)
        assert_false('FOO.AXIS.' in h)
        assert_false('FOO.' in h)
        assert_raises(KeyError, lambda: h['FOO.AXIS'])
        assert_raises(KeyError, lambda: h['FOO.AXIS.'])
        assert_raises(KeyError, lambda: h['FOO.'])



import __builtin__
import functools
import itertools
import mmap
import os
import signal
import sys
import tempfile
import textwrap
import threading
import warnings

try:
    try:
        from cStringIO import StringIO
    except ImportError:
        from StringIO import StringIO
except ImportError:
    from io import StringIO

try:
    from io import BytesIO
except ImportError:
    BytesIO = StringIO

import numpy as np


BLOCK_SIZE = 2880  # the FITS block size


def itersubclasses(cls, _seen=None):
    """
    itersubclasses(cls)

    Generator over all subclasses of a given class, in depth first order.

    >>> list(itersubclasses(int)) == [bool]
    True
    >>> class A(object): pass
    >>> class B(A): pass
    >>> class C(A): pass
    >>> class D(B,C): pass
    >>> class E(D): pass
    >>>
    >>> for cls in itersubclasses(A):
    ...     print(cls.__name__)
    B
    D
    E
    C
    >>> # get ALL (new-style) classes currently defined
    >>> [cls.__name__ for cls in itersubclasses(object)] #doctest: +ELLIPSIS
    ['type', ...'tuple', ...]

    From http://code.activestate.com/recipes/576949/
    """

    if not isinstance(cls, type):
        raise TypeError('itersubclasses must be called with '
                        'new-style classes, not %.100r' % cls)
    if _seen is None:
        _seen = set()
    try:
        subs = cls.__subclasses__()
    except TypeError:  # fails only when cls is type
        subs = cls.__subclasses__(cls)
    for sub in subs:
        if sub not in _seen:
            _seen.add(sub)
            yield sub
            for sub in itersubclasses(sub, _seen):
                yield sub


class lazyproperty(object):
    """
    Works similarly to property(), but computes the value only once.

    Adapted from the recipe at
    http://code.activestate.com/recipes/363602-lazy-property-evaluation
    """

    def __init__(self, fget, fset=None, fdel=None, doc=None):
        self._fget = fget
        self._fset = fset
        self._fdel = fdel
        if doc is None:
            self.__doc__ = fget.__doc__
        else:
            self.__doc__ = doc

    def __get__(self, obj, owner=None):
        if obj is None:
            return self
        key = self._fget.func_name
        if key not in obj.__dict__:
            val = self._fget(obj)
            obj.__dict__[key] = val
            return val
        else:
            return obj.__dict__[key]

    def __set__(self, obj, val):
        if self._fset:
            self._fset(obj, val)
        obj.__dict__[self._fget.func_name] = val

    def __delete__(self, obj):
        if self._fdel:
            self._fdel(obj)
        key = self._fget.func_name
        if key in obj.__dict__:
            del obj.__dict__[key]

    def getter(self, fget):
        return self.__ter(fget, 0)

    def setter(self, fset):
        return self.__ter(fset, 1)

    def deleter(self, fdel):
        return self.__ter(fdel, 2)

    def __ter(self, f, arg):
        args = [self._fget, self._fset, self._fdel, self.__doc__]
        args[arg] = f
        cls_ns = sys._getframe(1).f_locals
        for k, v in cls_ns.iteritems():
            if v is self:
                property_name = k
                break

        cls_ns[property_name] = lazyproperty(*args)

        return cls_ns[property_name]


# TODO: Provide a class deprecation marker as well.
def deprecated(since, message='', name='', alternative='', pending=False):
    """
    Used to mark a function as deprecated.

    To mark an attribute as deprecated, replace that attribute with a
    depcrecated property.

    Parameters
    ------------
    since : str
        The release at which this API became deprecated.  This is required.

    message : str, optional
        Override the default deprecation message.  The format specifier
        %(func)s may be used for the name of the function, and %(alternative)s
        may be used in the deprecation message to insert the name of an
        alternative to the deprecated function.

    name : str, optional
        The name of the deprecated function; if not provided the name is
        automatically determined from the passed in function, though this is
        useful in the case of renamed functions, where the new function is just
        assigned to the name of the deprecated function.  For example:
            def new_function():
                ...
            oldFunction = new_function

    alternative : str, optional
        An alternative function that the user may use in place of the
        deprecated function.  The deprecation warning will tell the user about
        this alternative if provided.

    pending : bool, optional
        If True, uses a PendingDeprecationWarning instead of a
        DeprecationWarning.

    """

    def deprecate(func, message=message, name=name, alternative=alternative,
                  pending=pending):
        if isinstance(func, classmethod):
            try:
                func = func.__func__
            except AttributeError:
                # classmethods in Python2.6 and below lack the __func__
                # attribute so we need to hack around to get it
                method = func.__get__(None, object)
                if hasattr(method, '__func__'):
                    func = method.__func__
                elif hasattr(method, 'im_func'):
                    func = method.im_func
                else:
                    # Nothing we can do really...  just return the original
                    # classmethod
                    return func
            is_classmethod = True
        else:
            is_classmethod = False

        if not name:
            name = func.__name__

        altmessage = ''
        if not message or type(message) == type(deprecate):
            if pending:
                message = ('The %(func)s function will be deprecated in a '
                           'future version.')
            else:
                message = ('The %(func)s function is deprecated and may '
                           'be removed in a future version.')
            if alternative:
                altmessage = '\n        Use %s instead.' % alternative

        message = ((message % {'func': name, 'alternative': alternative}) +
                   altmessage)

        @functools.wraps(func)
        def deprecated_func(*args, **kwargs):
            if pending:
                category = PendingDeprecationWarning
            else:
                category = DeprecationWarning

            warnings.warn(message, category, stacklevel=2)

            return func(*args, **kwargs)

        old_doc = deprecated_func.__doc__
        if not old_doc:
            old_doc = ''
        old_doc = textwrap.dedent(old_doc).strip('\n')
        altmessage = altmessage.strip()
        if not altmessage:
            altmessage = message.strip()
        new_doc = (('\n.. deprecated:: %(since)s'
                    '\n    %(message)s\n\n' %
                    {'since': since, 'message': altmessage.strip()}) + old_doc)
        if not old_doc:
            # This is to prevent a spurious 'unexected unindent' warning from
            # docutils when the original docstring was blank.
            new_doc += r'\ '

        deprecated_func.__doc__ = new_doc

        if is_classmethod:
            deprecated_func = classmethod(deprecated_func)
        return deprecated_func

    if type(message) == type(deprecate):
        return deprecate(message)

    return deprecate


def ignore_sigint(func):
    """
    This decorator registers a custom SIGINT handler to catch and ignore SIGINT
    until the wrapped function is completed.
    """

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        # Get the name of the current thread and determine if this is a single
        # treaded application
        curr_thread = threading.currentThread()
        single_thread = (threading.activeCount() == 1 and
                         curr_thread.getName() == 'MainThread')

        class SigintHandler(object):
            def __init__(self):
                self.sigint_received = False

            def __call__(self, signum, frame):
                warnings.warn('KeyboardInterrupt ignored until %s is '
                              'complete!' % func.__name__)
                self.sigint_received = True

        sigint_handler = SigintHandler()

        # Define new signal interput handler
        if single_thread:
            # Install new handler
            old_handler = signal.signal(signal.SIGINT, sigint_handler)

        try:
            func(*args, **kwargs)
        finally:
            if single_thread:
                if old_handler is not None:
                    signal.signal(signal.SIGINT, old_handler)
                else:
                    signal.signal(signal.SIGINT, signal.SIG_DFL)

                if sigint_handler.sigint_received:
                    raise KeyboardInterrupt

    return wrapped


def pairwise(iterable):
    """Return the items of an iterable paired with its next item.

    Ex: s -> (s0,s1), (s1,s2), (s2,s3), ....
    """

    a, b = itertools.tee(iterable)
    for _ in b:
        # Just a little trick to advance b without having to catch
        # StopIter if b happens to be empty
        break
    return itertools.izip(a, b)


def isiterable(obj):
    """Returns true of the given object is iterable."""

    # In Python2.6 and up this is simply a matter of checking isinstance
    # collections.Iterable, but this unavailable in Python 2.5 and below
    try:
        from collections import Iterable
        if isinstance(obj, Iterable):
            return True
    except ImportError:
        pass

    try:
        iter(obj)
        return True
    except TypeError:
        return False


def encode_ascii(s):
    """
    In Python 2 this is a no-op.  Strings are left alone.  In Python 3 this
    will be replaced with a function that actually encodes unicode strings to
    ASCII bytes.
    """

    return s


def decode_ascii(s):
    """
    In Python 2 this is a no-op.  Strings are left alone.  In Python 3 this
    will be replaced with a function that actually decodes ascii bytes to
    unicode.
    """

    return s


def isreadable(f):
    """
    Returns True if the file-like object can be read from.  This is a common-
    sense approximation of io.IOBase.readable.
    """

    if hasattr(f, 'closed') and f.closed:
        # This mimics the behavior of io.IOBase.readable
        raise ValueError('I/O operation on closed file')

    if not hasattr(f, 'read'):
        return False

    if hasattr(f, 'mode') and not any((c in f.mode for c in 'r+')):
        return False

    # Not closed, has a 'read()' method, and either has no known mode or a
    # readable mode--should be good enough to assume 'readable'
    return True


def iswritable(f):
    """
    Returns True if the file-like object can be written to.  This is a common-
    sense approximation of io.IOBase.writable.
    """

    if hasattr(f, 'closed') and f.closed:
        # This mimics the behavior of io.IOBase.writable
        raise ValueError('I/O operation on closed file')

    if not hasattr(f, 'write'):
        return False

    if hasattr(f, 'mode') and not any((c in f.mode for c in 'wa+')):
        return False

    # Note closed, has a 'write()' method, and either has no known mode or a
    # mode that supports writing--should be good enough to assume 'writable'
    return True


def isfile(f):
    """
    Returns True if the given object represents an OS-level file (that is,
    isinstance(f, file)).

    On Python 3 this also returns True if the given object is higher level
    wrapper on top of a FileIO object, such as a TextIOWrapper.
    """

    return isinstance(f, file)


def fileobj_open(filename, mode):
    """
    A wrapper around the `open()` builtin.

    This exists because in Python 3, `open()` returns an `io.BufferedReader` by
    default.  This is bad, because `io.BufferedReader` doesn't support random
    access, which we need in some cases.  In the Python 3 case (implemented in
    the py3compat module) we must call open with buffering=0 to get a raw
    random-access file reader.
    """

    return open(filename, mode)


def fileobj_name(f):
    """
    Returns the 'name' of file-like object f, if it has anything that could be
    called its name.  Otherwise f's class or type is returned.  If f is a
    string f itself is returned.
    """

    if isinstance(f, basestring):
        return f
    elif hasattr(f, 'name'):
        return f.name
    elif hasattr(f, 'filename'):
        return f.filename
    elif hasattr(f, '__class__'):
        return str(f.__class__)
    else:
        return str(type(f))


def fileobj_closed(f):
    """
    Returns True if the given file-like object is closed or if f is not a
    file-like object.
    """

    if hasattr(f, 'closed'):
        return f.closed
    elif hasattr(f, 'fileobj') and hasattr(f.fileobj, 'closed'):
        return f.fileobj.closed
    elif hasattr(f, 'fp') and hasattr(f.fp.closed):
        return f.fp.closed
    else:
        return True


def fileobj_mode(f):
    """
    Returns the 'mode' string of a file-like object if such a thing exists.
    Otherwise returns None.
    """

    # Go from most to least specific--for example gzip objects have a 'mode'
    # attribute, but it's not analogous to the file.mode attribute
    if hasattr(f, 'fileobj') and hasattr(f.fileobj, 'mode'):
        mode = f.fileobj.mode
    elif hasattr(f, 'fp') and hasattr(f.fp, 'mode'):
        mode = f.fp.mode
    elif hasattr(f, 'mode'):
        mode = f.mode
    else:
        mode = None

    if isinstance(mode, basestring):
        # On Python 3 files opened in 'a' mode actually get opened in 'w' or
        # 'r+' instead of 'a', since the behavior of 'a' mode is not normalized
        # across systems.  So we should normalize in the same way...
        if 'a' in mode:
            if '+' in mode:
                mode = mode.replace('a', 'r')
            else:
                mode = mode.replace('a', 'w')
    else:
        # If mode still turned out to be something other than a string, it's
        # not the droid we were looking for, and we should just toss it
        mode = None
    return mode


def fileobj_is_binary(f):
    """
    Returns True if the give file or file-like object has a file open in binary
    mode.  When in doubt, returns True by default.
    """

    # TODO: In Python 3 it might be more reliable to check if the fileobj is a
    # text reader or a binary reader

    mode = fileobj_mode(f)
    if mode:
        return 'b' in mode
    else:
        return True


def translate(s, table, deletechars):
    """
    This is a version of string/unicode.translate() that can handle string or
    unicode strings the same way using a translation table made with
    string.maketrans.
    """

    if isinstance(s, str):
        return s.translate(table, deletechars)
    elif isinstance(s, unicode):
        table = dict((x, ord(table[x])) for x in range(256)
                     if ord(table[x]) != x)
        for c in deletechars:
            table[ord(c)] = None
        return s.translate(table)


def indent(s, shift=1, width=4):
    indented = '\n'.join(' ' * (width * shift) + l if l else ''
                         for l in s.splitlines())
    if s[-1] == '\n':
        indented += '\n'

    return indented


def fill(text, width, *args, **kwargs):
    """
    Like :func:`textwrap.wrap` but preserves existing paragraphs which
    :func:`textwrap.wrap` does not otherwise handle well.  Also handles section
    headers.
    """

    paragraphs = text.split('\n\n')
    def maybe_fill(t):
        if all(len(l) < width for l in t.splitlines()):
            return t
        else:
            return textwrap.fill(t, width, *args, **kwargs)

    return '\n\n'.join(maybe_fill(p) for p in paragraphs)


def _array_from_file(infile, dtype, count, sep):
    """Create a numpy array from a file or a file-like object."""

    if isfile(infile):
        return np.fromfile(infile, dtype=dtype, count=count, sep=sep)
    else:
        # treat as file-like object with "read" method; this includes gzip file
        # objects, because numpy.fromfile just reads the compressed bytes from
        # their underlying file object, instead of the decompresed bytes
        read_size = np.dtype(dtype).itemsize * count
        s = infile.read(read_size)
        return np.fromstring(s, dtype=dtype, count=count, sep=sep)


def _array_to_file(arr, outfile):
    """Write a numpy array to a file or a file-like object."""

    if isfile(outfile):
        arr.tofile(outfile)
    else:
        # treat as file-like object with "write" method
        _write_string(outfile, arr.tostring())


def _write_string(f, s):
    """
    Write a string to a file, encoding to ASCII if the file is open in binary
    mode, or decoding if the file is open in text mode.
    """

    # Assume if the file object doesn't have a specific mode, that the mode is
    # binary
    binmode = fileobj_is_binary(f)

    if binmode and isinstance(s, unicode):
        s = encode_ascii(s)
    elif not binmode and not isinstance(f, unicode):
        s = decode_ascii(s)
    f.write(s)


def _convert_array(array, dtype):
    """
    Converts an array to a new dtype--if the itemsize of the new dtype is
    the same as the old dtype and both types are not numeric, a view is
    returned.  Otherwise a new array must be created.
    """

    if array.dtype == dtype:
        return array
    elif (array.dtype.itemsize == dtype.itemsize and not
            (np.issubdtype(array.dtype, np.number) and
             np.issubdtype(dtype, np.number))):
        # Includes a special case when both dtypes are at least numeric to
        # account for ticket #218: https://trac.assembla.com/pyfits/ticket/218
        return array.view(dtype)
    else:
        return array.astype(dtype)


def _unsigned_zero(dtype):
    """
    Given a numpy dtype, finds its "zero" point, which is exactly in the
    middle of its range.
    """

    assert dtype.kind == 'u'
    return 1 << (dtype.itemsize * 8 - 1)


def _is_pseudo_unsigned(dtype):
    return dtype.kind == 'u' and dtype.itemsize >= 2


def _is_int(val):
    return isinstance(val, (int, long, np.integer))


def _str_to_num(val):
    """Converts a given string to either an int or a float if necessary."""

    try:
        num = int(val)
    except ValueError:
        # If this fails then an exception should be raised anyways
        num = float(val)
    return num


def _pad_length(stringlen):
    """Bytes needed to pad the input stringlen to the next FITS block."""

    return (BLOCK_SIZE - (stringlen % BLOCK_SIZE)) % BLOCK_SIZE


def _normalize_slice(input, naxis):
    """
    Set the slice's start/stop in the regular range.
    """

    def _normalize(indx, npts):
        if indx < -npts:
            indx = 0
        elif indx < 0:
            indx += npts
        elif indx > npts:
            indx = npts
        return indx

    _start = input.start
    if _start is None:
        _start = 0
    elif _is_int(_start):
        _start = _normalize(_start, naxis)
    else:
        raise IndexError('Illegal slice %s; start must be integer.' % input)

    _stop = input.stop
    if _stop is None:
        _stop = naxis
    elif _is_int(_stop):
        _stop = _normalize(_stop, naxis)
    else:
        raise IndexError('Illegal slice %s; stop must be integer.' % input)

    if _stop < _start:
        raise IndexError('Illegal slice %s; stop < start.' % input)

    _step = input.step
    if _step is None:
        _step = 1
    elif _is_int(_step):
        if _step <= 0:
            raise IndexError('Illegal slice %s; step must be positive.'
                             % input)
    else:
        raise IndexError('Illegal slice %s; step must be integer.' % input)

    return slice(_start, _stop, _step)


def _words_group(input, strlen):
    """
    Split a long string into parts where each part is no longer
    than `strlen` and no word is cut into two pieces.  But if
    there is one single word which is longer than `strlen`, then
    it will be split in the middle of the word.
    """

    words = []
    nblanks = input.count(' ')
    nmax = max(nblanks, len(input) // strlen + 1)
    arr = np.fromstring((input + ' '), dtype=(bytes, 1))

    # locations of the blanks
    blank_loc = np.nonzero(arr == ' '.encode('latin1'))[0]
    offset = 0
    xoffset = 0
    for idx in range(nmax):
        try:
            loc = np.nonzero(blank_loc >= strlen + offset)[0][0]
            offset = blank_loc[loc - 1] + 1
            if loc == 0:
                offset = -1
        except:
            offset = len(input)

        # check for one word longer than strlen, break in the middle
        if offset <= xoffset:
            offset = xoffset + strlen

        # collect the pieces in a list
        words.append(input[xoffset:offset])
        if len(input) == offset:
            break
        xoffset = offset

    return words


def _tmp_name(input):
    """
    Create a temporary file name which should not already exist.  Use the
    directory of the input file as the base name of the mkstemp() output.
    """

    if input is not None:
        input = os.path.dirname(input)
    f, fn = tempfile.mkstemp(dir=input)
    os.close(f)
    return fn


if sys.version_info[:2] < (2, 6):
    # In Python 2.5 mmap.mmap is a function that returns an object of type
    # 'mmap.mmap', but the mmap.mmap type is otherwise not accessible through
    # the module
    def _is_mmap(obj):
        return (type(obj).__module__ == 'mmap' and
                type(obj).__name__ == 'mmap')
else:
    def _is_mmap(obj):
        return isinstance(obj, mmap.mmap)


def _get_array_mmap(array):
    """
    If the array has an mmap.mmap at base of its base chain, return the mmap
    object; otherwise return None.
    """

    if _is_mmap(array):
        return array

    base = array
    while hasattr(base, 'base') and base.base is not None:
        if _is_mmap(base.base):
            return base.base
        base = base.base


if sys.version_info[:2] < (2, 6):
    import __builtin__
    # Replace the builtin property to add support for the getter/setter/deleter
    # mechanism as introduced in Python 2.6 (this can go away if we ever drop
    # 2.5 support)
    class property(property):
        def __init__(self, fget, *args, **kwargs):
            self.__doc__ = fget.__doc__
            super(property, self).__init__(fget, *args, **kwargs)

        def getter(self, fget):
            return self.__ter(fget, 0)

        def setter(self, fset):
            return self.__ter(fset, 1)

        def deleter(self, fdel):
            return self.__ter(fdel, 2)

        def __ter(self, f, arg):
            args = [self.fget, self.fset, self.fdel, self.__doc__]
            args[arg] = f
            cls_ns = sys._getframe(1).f_locals
            for k, v in cls_ns.iteritems():
                if v is self:
                    property_name = k
                    break

            cls_ns[property_name] = property(*args)

            return cls_ns[property_name]
    __builtin__.property = property



import operator
import sys
import warnings
import weakref

import numpy as np

from pyfits.column import (ASCIITNULL, FITS2NUMPY, Column, ColDefs, _FormatX,
                           _FormatP, _VLF, _get_index, _wrapx, _unwrapx,
                            _convert_ascii_format)
from pyfits.util import _array_from_file, decode_ascii, lazyproperty


class FITS_record(object):
    """
    FITS record class.

    `FITS_record` is used to access records of the `FITS_rec` object.
    This will allow us to deal with scaled columns.  It also handles
    conversion/scaling of columns in ASCII tables.  The `FITS_record`
    class expects a `FITS_rec` object as input.
    """

    def __init__(self, input, row=0, start=None, end=None, step=None,
                 base=None, **kwargs):
        """
        Parameters
        ----------
        input : array
           The array to wrap.

        row : int, optional
           The starting logical row of the array.

        start : int, optional
           The starting column in the row associated with this object.
           Used for subsetting the columns of the FITS_rec object.

        end : int, optional
           The ending column in the row associated with this object.
           Used for subsetting the columns of the FITS_rec object.
        """

        # For backward compatibility...
        for arg in [('startColumn', 'start'), ('endColumn', 'end')]:
            if arg[0] in kwargs:
                warnings.warn('The %s argument to FITS_record is deprecated; '
                              'use %s instead' % arg, DeprecationWarning)
                if arg[0] == 'startColumn':
                    start = kwargs[arg[0]]
                elif arg[0] == 'endColumn':
                    end = kwargs[arg[0]]

        self.array = input
        self.row = row
        if base:
            width = len(base)
        else:
            width = self.array._nfields

        s = slice(start, end, step).indices(width)
        self.start, self.end, self.step = s
        self.base = base

    def __getitem__(self, key):
        if isinstance(key, basestring):
            indx = _get_index(self.array.names, key)

            if indx < self.start or indx > self.end - 1:
                raise KeyError("Key '%s' does not exist." % key)
        elif isinstance(key, slice):
            return type(self)(self.array, self.row, key.start, key.stop,
                               key.step, self)
        else:
            indx = self._get_index(key)

            if indx > self.array._nfields - 1:
                raise IndexError('Index out of bounds')

        return self.array.field(indx)[self.row]

    def __setitem__(self, key, value):
        if isinstance(key, basestring):
            indx = _get_index(self.array._coldefs.names, key)

            if indx < self.start or indx > self.end - 1:
                raise KeyError("Key '%s' does not exist." % key)
        elif isinstance(key, slice):
            for indx in xrange(slice.start, slice.stop, slice.step):
                indx = self._get_indx(indx)
                self.array.field(indx)[self.row] = value
        else:
            indx = self._get_index(key)
            if indx > self.array._nfields - 1:
                raise IndexError('Index out of bounds')

        self.array.field(indx)[self.row] = value

    def __getslice__(self, start, end):
        return self[slice(start, end)]

    def __len__(self):
        return len(xrange(self.start, self.end, self.step))

    def __repr__(self):
        """
        Display a single row.
        """

        outlist = []
        for idx in xrange(len(self)):
            outlist.append(repr(self[idx]))
        return '(%s)' % ', '.join(outlist)

    def field(self, field):
        """
        Get the field data of the record.
        """

        return self.__getitem__(field)

    def setfield(self, field, value):
        """
        Set the field data of the record.
        """

        self.__setitem__(field, value)

    @lazyproperty
    def _bases(self):
        bases = [weakref.proxy(self)]
        base = self.base
        while base:
            bases.append(base)
            base = base.base
        return bases

    def _get_index(self, indx):
        indices = np.ogrid[:self.array._nfields]
        for base in reversed(self._bases):
            if base.step < 1:
                s = slice(base.start, None, base.step)
            else:
                s = slice(base.start, base.end, base.step)
            indices = indices[s]
        return indices[indx]


class FITS_rec(np.recarray):
    """
    FITS record array class.

    `FITS_rec` is the data part of a table HDU's data part.  This is a
    layer over the `recarray`, so we can deal with scaled columns.

    It inherits all of the standard methods from `numpy.ndarray`.
    """

    _record_type = FITS_record

    def __new__(subtype, input):
        """
        Construct a FITS record array from a recarray.
        """

        # input should be a record array
        if input.dtype.subdtype is None:
            self = np.recarray.__new__(subtype, input.shape, input.dtype,
                                       buf=input.data)
        else:
            self = np.recarray.__new__(subtype, input.shape, input.dtype,
                                       buf=input.data, strides=input.strides)

        self._nfields = len(self.dtype.names)
        self._convert = [None] * len(self.dtype.names)
        self._heapoffset = 0
        self._heapsize = 0
        self._file = None
        self._buffer = None
        self._coldefs = None
        self._gap = 0
        self.names = list(self.dtype.names)
        self.formats = None
        return self

    def __array_finalize__(self, obj):
        if obj is None:
            return

        if isinstance(obj, FITS_rec):
            self._convert = obj._convert
            self._heapoffset = obj._heapoffset
            self._heapsize = obj._heapsize
            self._file = obj._file
            self._buffer = obj._buffer
            self._coldefs = obj._coldefs
            self._nfields = obj._nfields
            self._gap = obj._gap
            self.names = obj.names
            self.formats = obj.formats
        else:
            # This will allow regular ndarrays with fields, rather than
            # just other FITS_rec objects
            self._nfields = len(obj.dtype.names)
            self._convert = [None] * len(obj.dtype.names)

            self._heapoffset = getattr(obj, '_heapoffset', 0)
            self._heapsize = getattr(obj, '_heapsize', 0)
            self._file = getattr(obj, '_file', None)
            self._buffer = getattr(obj, '_buffer', None)

            self._coldefs = None
            self._gap = 0

            # Bypass setattr-based assignment to fields; see #86
            self.names = list(obj.dtype.names)
            self.formats = None

            attrs = ['_convert', '_coldefs', '_gap']
            for attr in attrs:
                if hasattr(obj, attr):
                    value = getattr(obj, attr, None)
                    if value is None:
                        warnings.warn('Setting attribute %s as None' % attr)
                    setattr(self, attr, value)

            if self._coldefs is None:
                self._coldefs = ColDefs(self)
            self.formats = self._coldefs.formats

    def __repr__(self):
        return np.recarray.__repr__(self)

    def __getattribute__(self, attr):
        # See the comment in __setattr__
        if attr in ('names', 'formats'):
            return object.__getattribute__(self, attr)
        else:
            return super(FITS_rec, self).__getattribute__(attr)

    def __setattr__(self, attr, value):
        # Overrides the silly attribute-based assignment to fields supported by
        # recarrays for our two built-in public attributes: names and formats
        # Otherwise, the default behavior, bad as it is, is preserved.  See
        # ticket #86
        if attr in ('names', 'formats'):
            return object.__setattr__(self, attr, value)
        else:
            return super(FITS_rec, self).__setattr__(attr, value)

    def __getitem__(self, key):
        if isinstance(key, basestring):
            return self.field(key)
        elif isinstance(key, (slice, np.ndarray, tuple, list)):
            # Have to view as a recarray then back as a FITS_rec, otherwise the
            # circular reference fix/hack in FITS_rec.field() won't preserve
            # the slice
            subtype = type(self)
            out = self.view(np.recarray).__getitem__(key).view(subtype)
            out._coldefs = ColDefs(self._coldefs)
            arrays = []
            out._convert = [None] * len(self.dtype.names)
            for idx in range(len(self.dtype.names)):
                #
                # Store the new arrays for the _coldefs object
                #
                arrays.append(self._coldefs._arrays[idx][key])

                # touch all fields to expand the original ._convert list
                # so the sliced FITS_rec will view the same scaled columns as
                # the original
                dummy = self.field(idx)
                if self._convert[idx] is not None:
                    out._convert[idx] = \
                        np.ndarray.__getitem__(self._convert[idx], key)
            del dummy

            out._coldefs._arrays = arrays
            out._coldefs._shape = len(arrays[0])

            return out

        # if not a slice, do this because Record has no __getstate__.
        # also more efficient.
        else:
            if isinstance(key, int) and key >= len(self):
                raise IndexError("Index out of bounds")

            newrecord = self._record_type(self, key)
            return newrecord

    def __setitem__(self, row, value):
        if isinstance(row, slice):
            end = min(len(self), row.stop or len(self))
            end = max(0, end)
            start = max(0, row.start or 0)
            end = min(end, start + len(value))

            for idx in range(start, end):
                self.__setitem__(idx, value[idx - start])
            return

        if isinstance(value, FITS_record):
            for idx in range(self._nfields):
                self.field(self.names[idx])[row] = value.field(self.names[idx])
        elif isinstance(value, (tuple, list)):
            if self._nfields == len(value):
                for idx in range(self._nfields):
                    self.field(idx)[row] = value[idx]
            else:
                raise ValueError('Input tuple or list required to have %s '
                                 'elements.' % self._nfields)
        else:
            raise TypeError('Assignment requires a FITS_record, tuple, or '
                            'list as input.')

    def __getslice__(self, start, end):
        return self[slice(start, end)]

    def __setslice__(self, start, end, value):
        self[slice(start, end)] = value

    @property
    def columns(self):
        """
        A user-visible accessor for the coldefs.  See ticket #44.
        """

        return self._coldefs

    def field(self, key):
        """
        A view of a `Column`'s data as an array.
        """

        indx = _get_index(self.names, key)
        recformat = self._coldefs._recformats[indx]

        # If field's base is a FITS_rec, we can run into trouble because it
        # contains a reference to the ._coldefs object of the original data;
        # this can lead to a circular reference; see ticket #49
        base = self
        while isinstance(base, FITS_rec) and \
              isinstance(base.base, np.recarray):
            base = base.base
        # base could still be a FITS_rec in some cases, so take care to
        # use rec.recarray.field to avoid a potential infinite
        # recursion
        field = np.recarray.field(base, indx)

        if self._convert[indx] is None:
            # for X format
            if isinstance(recformat, _FormatX):
                _nx = recformat._nx
                dummy = np.zeros(self.shape + (_nx,), dtype=np.bool_)
                _unwrapx(field, dummy, _nx)
                self._convert[indx] = dummy
                return self._convert[indx]

            (_str, _bool, _number, _scale, _zero, bscale, bzero, dim) = \
                self._get_scale_factors(indx)

            # for P format
            if isinstance(recformat, _FormatP):
                dummy = _VLF([None] * len(self), dtype=recformat.dtype)
                raw_data = self._get_raw_data()
                if raw_data is None:
                    raise IOError(
                        "Could not find heap data for the %r variable-length "
                        "array column." % self.names[indx])
                for i in range(len(self)):
                    offset = field[i, 1] + self._heapoffset
                    count = field[i, 0]

                    if recformat.dtype == 'a':
                        dt = np.dtype(recformat.dtype + str(1))
                        arr_len = count * dt.itemsize
                        da = raw_data[offset:offset + arr_len].view(dt)
                        da = np.char.array(da.view(dtype=dt), itemsize=count)
                        dummy[i] = decode_ascii(da)
                    else:
                        dt = np.dtype(recformat.dtype)
                        arr_len = count * dt.itemsize
                        dummy[i] = raw_data[offset:offset + arr_len].view(dt)
                        dummy[i].dtype = dummy[i].dtype.newbyteorder('>')

                # scale by TSCAL and TZERO
                if _scale or _zero:
                    for i in range(len(self)):
                        dummy[i][:] = dummy[i] * bscale + bzero

                # Boolean (logical) column
                if recformat.dtype == FITS2NUMPY['L']:
                    for i in range(len(self)):
                        dummy[i] = np.equal(dummy[i], ord('T'))

                self._convert[indx] = dummy
                return self._convert[indx]

            # ASCII table, convert strings to numbers
            if not _str and self._coldefs._tbtype == 'TableHDU':
                _fmap = {'I': np.int32, 'F': np.float32, 'E': np.float32,
                         'D': np.float64}
                _type = _fmap[self._coldefs.formats[indx][0]]

                # if the string = TNULL, return ASCIITNULL
                nullval = self._coldefs.nulls[indx].strip().encode('ascii')
                dummy = field.replace('D'.encode('ascii'),
                                      'E'.encode('ascii'))
                dummy = np.where(dummy.strip() == nullval, str(ASCIITNULL),
                                 dummy)
                try:
                    dummy = np.array(dummy, dtype=_type)
                except ValueError, e:
                    raise ValueError(
                        '%s; the header may be missing the necessary TNULL%d '
                        'keyword or the table contains invalid data' %
                        (e, indx + 1))

                self._convert[indx] = dummy
            else:
                dummy = field

            # Test that the dimensions given in dim are sensible; otherwise
            # display a warning and ignore them
            if dim:
                # See if the dimensions already match, if not, make sure the
                # number items will fit in the specified dimensions
                if dummy.ndim > 1:
                    actual_shape = dummy[0].shape
                    if _str:
                        actual_shape = (dummy[0].itemsize,) + actual_shape
                else:
                    actual_shape = len(dummy[0])
                if dim == actual_shape:
                    # The array already has the correct dimensions, so we
                    # ignore dim and don't convert
                    dim = None
                else:
                    nitems = reduce(operator.mul, dim)
                    if _str:
                        actual_nitems = dummy.itemsize
                    else:
                        actual_nitems = dummy.shape[1]
                    if nitems > actual_nitems:
                        warnings.warn(
                        'TDIM%d value %s does not fit with the size of '
                            'the array items (%d).  TDIM%d will be ignored.'
                            % (indx + 1, self._coldefs.dims[indx],
                               actual_nitems, indx + 1))
                        dim = None

            # further conversion for both ASCII and binary tables
            if _number and (_scale or _zero):

                # only do the scaling the first time and store it in _convert
                self._convert[indx] = np.array(dummy, dtype=np.float64)
                if _scale:
                    np.multiply(self._convert[indx], bscale,
                                self._convert[indx])
                if _zero:
                    self._convert[indx] += bzero
            elif _bool and dummy.dtype != bool:
                self._convert[indx] = np.equal(dummy, ord('T'))
            elif _str:
                try:
                    self._convert[indx] = decode_ascii(dummy)
                except UnicodeDecodeError:
                    pass

            if dim:
                nitems = reduce(operator.mul, dim)
                if self._convert[indx] is None:
                    self._convert[indx] = dummy[:,:nitems]
                if _str:
                    fmt = self._convert[indx].dtype.char
                    dtype = ('|%s%d' % (fmt, dim[-1]), dim[:-1])
                    self._convert[indx].dtype = dtype
                else:
                    self._convert[indx].shape = (dummy.shape[0],) + dim

        if self._convert[indx] is not None:
            return self._convert[indx]
        else:
            return dummy

    def _clone(self, shape):
        """
        Overload this to make mask array indexing work properly.
        """

        from pyfits.hdu.table import new_table

        hdu = new_table(self._coldefs, nrows=shape[0])
        return hdu.data

    def _get_raw_data(self):
        """
        Returns the base array of self that "raw data array" that is the
        array in the format that it was first read from a file before it was
        sliced or viewed as a different type in any way.

        This is determined by walking through the bases until finding one that
        has at least the same number of bytes as self, plus the heapsize.  This
        may be the immediate .base but is not always.  This is used primarily
        for variable-length array support which needs to be able to find the
        heap (the raw data *may* be larger than nbytes + heapsize if it
        contains a gap or padding).

        May return ``None`` if no array resembling the "raw data" according to
        the stated criteria can be found.
        """

        raw_data_bytes = self.nbytes + self._heapsize
        base = self
        while hasattr(base, 'base') and base.base is not None:
            base = base.base
            if hasattr(base, 'nbytes') and base.nbytes >= raw_data_bytes:
                return base

    def _get_scale_factors(self, indx):
        """
        Get the scaling flags and factors for one field.

        `indx` is the index of the field.
        """

        if self._coldefs._tbtype == 'BinTableHDU':
            _str = 'a' in self._coldefs._recformats[indx]
            _bool = self._coldefs._recformats[indx][-2:] == FITS2NUMPY['L']
        else:
            _str = self._coldefs.formats[indx][0] == 'A'
            _bool = False             # there is no boolean in ASCII table
        _number = not(_bool or _str)
        bscale = self._coldefs.bscales[indx]
        bzero = self._coldefs.bzeros[indx]
        _scale = bscale not in ['', None, 1]
        _zero = bzero not in ['', None, 0]
        # ensure bscale/bzero are numbers
        if not _scale:
            bscale = 1
        if not _zero:
            bzero = 0

        dim = self._coldefs._dims[indx]

        return (_str, _bool, _number, _scale, _zero, bscale, bzero, dim)

    def _scale_back(self):
        """
        Update the parent array, using the (latest) scaled array.
        """

        _fmap = {'A': 's', 'I': 'd', 'J': 'd', 'F': 'f', 'E': 'E', 'D': 'E'}
        # calculate the starting point and width of each field for ASCII table
        # TODO: Ick--fix this _tbtype usage eventually...
        if self._coldefs._tbtype == 'TableHDU':
            loc = self._coldefs.starts
            widths = []

            idx = 0
            for idx in range(len(self.dtype.names)):
                f = _convert_ascii_format(self._coldefs.formats[idx])
                widths.append(f[1])
            loc.append(loc[-1] + super(FITS_rec, self).field(idx).itemsize)

        for indx in range(len(self.dtype.names)):
            recformat = self._coldefs._recformats[indx]
            field = super(FITS_rec, self).field(indx)

            if self._convert[indx] is None:
                continue

            if isinstance(recformat, _FormatX):
                _wrapx(self._convert[indx], field, recformat._nx)
                continue

            (_str, _bool, _number, _scale, _zero, bscale, bzero, dim) = \
                self._get_scale_factors(indx)

            # add the location offset of the heap area for each
            # variable length column
            if isinstance(recformat, _FormatP):
                # Reset the heapsize and recompute it starting from the first P
                # column
                if indx == 0:
                    self._heapsize = 0

                field[:] = 0  # reset
                npts = map(len, self._convert[indx])

                dtype = np.array([], dtype=recformat.dtype).dtype
                field[:len(npts), 0] = npts
                field[1:, 1] = (np.add.accumulate(field[:-1, 0]) *
                                dtype.itemsize)
                field[:, 1][:] += self._heapsize
                self._heapsize += field[:, 0].sum() * dtype.itemsize

            # conversion for both ASCII and binary tables
            if _number or _str:
                if _number and (_scale or _zero):
                    dummy = self._convert[indx].copy()
                    if _zero:
                        dummy -= bzero
                    if _scale:
                        dummy /= bscale
                elif _str:
                    dummy = self._convert[indx]
                elif self._coldefs._tbtype == 'TableHDU':
                    dummy = self._convert[indx]
                else:
                    continue

                # ASCII table, convert numbers to strings
                if self._coldefs._tbtype == 'TableHDU':
                    format = self._coldefs.formats[indx].strip()
                    lead = self._coldefs.starts[indx] - loc[indx]
                    if lead < 0:
                        raise ValueError(
                            'Column `%s` starting point overlaps to the '
                            'previous column.' % indx + 1)
                    trail = (loc[indx + 1] - widths[indx] -
                             self._coldefs.starts[indx])
                    if trail < 0:
                        raise ValueError(
                            'Column `%s` ending point overlaps to the next '
                            'column.' % indx + 1)
                    if 'A' in format:
                        _pc = '%-'
                    else:
                        _pc = '%'

                    fmt = ''.join([(' ' * lead), _pc, format[1:],
                                   _fmap[format[0]], (' ' * trail)])

                    # not using numarray.strings's num2char because the
                    # result is not allowed to expand (as C/Python does).
                    for jdx in range(len(dummy)):
                        x = fmt % dummy[jdx]
                        if len(x) > (loc[indx + 1] - loc[indx]):
                            raise ValueError(
                                "Number `%s` does not fit into the output's "
                                "itemsize of %s." % (x, widths[indx]))
                        else:
                            field[jdx] = x
                    # Replace exponent separator in floating point numbers
                    if 'D' in format:
                        field.replace('E', 'D')
                # binary table
                else:
                    if len(field) and isinstance(field[0], np.integer):
                        dummy = np.around(dummy)
                    elif isinstance(field, np.chararray):
                        # Ensure that blanks at the end of each string are
                        # converted to nulls instead of spaces, see Trac #15
                        # and #111
                        itemsize = dummy.itemsize
                        if dummy.dtype.kind == 'U':
                            pad = self._coldefs._padding_byte
                        else:
                            pad = self._coldefs._padding_byte.encode('ascii')

                        for idx in xrange(len(dummy)):
                            val = dummy[idx]
                            dummy[idx] = val + (pad * (itemsize - len(val)))

                        # Encode *after* handling the padding byte or else
                        # Numpy will complain about trying to append bytes to
                        # an array
                        if dummy.dtype.kind == 'U':
                            dummy = dummy.encode('ascii')

                    if field.shape == dummy.shape:
                        field[:] = dummy
                    else:
                        # Reshaping the data is necessary in cases where the
                        # TDIMn keyword was used to shape a column's entries
                        # into arrays
                        field[:] = dummy.ravel().view(field.dtype)

                del dummy

            # ASCII table does not have Boolean type
            elif _bool:
                field[:] = np.choose(self._convert[indx],
                                     (np.array([ord('F')], dtype=np.int8)[0],
                                      np.array([ord('T')], dtype=np.int8)[0]))



"""
This subpackage contains implementations of command-line scripts that are
included with PyFITS.

It lives under the pyfits package primarily for Python 3 support--2to3 does not
convert these scripts unless they are actually found in the packages (as
opposed to the top-level scripts/ directory.

The actual scripts that are installed in bin/ are simple wrappers for these
modules that will run in any Python version.
"""
